#!/usr/bin/perl

=pod

KIND - The taxonomy project

    Copyright (C) 2020  Rogelio Nazar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.



~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Notes on versions:


February 24, 2021

We found a terrible bug with the extraction of contexts. 
It is now fixed.






December 14, 2020

We just jettisoned the embed module:
it was too complicated and added nothing to
the final classification. 

At some point later we will document this
ex-module (which will be renamed wembe, 
for word-embedding):

http://www.tecling.com/kind/wembe


~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



Documentation:

There is no documentation here. Refer to the documentation for 
a detailed explanation.
It is absolutely imperative to read the documentation 
before any attempt to make this thing work.

http://www.tecling.com/cgi-bin/kind/2020/index.pl?source=1

For more details, read our latest publication on the subject:

+ Nazar, R.; Balvet, A.; Ferraro, G.; Marín, R.; Renau, I. (2020). Pruning and repopulating a lexical taxonomy: experiments in Spanish, English and French. Journal of Intelligent Systems, vol. 30 num. 1, pp. 376-394.

https://www.degruyter.com/downloadpdf/journals/jisys/30/1/article-p376.xml

Here, there will only be some additional in-line commentary.

=cut

use CGI::Carp qw(fatalsToBrowser warningsToBrowser);
$|++;
use strict;
use CGI;
my $CGI = new CGI; 

our $self = $ENV{SCRIPT_NAME};
print $CGI->header(
-type => 'text/html',
-charset =>'utf-8',
 );

warningsToBrowser(1);

#########################
# Parameters
#########################
our $input = $CGI->param('input'); # the input can be a single noun 
				# or a new-line separated list of nouns
				# nouns can also be multi-word units.
$input = lc $input;
our $usecache = $CGI->param('usecache');
$usecache = 1 if (! defined($usecache));
our $ignorecache = $CGI->param('ignorecache');
$usecache = 0 if ($ignorecache); # you can force it to not-use the cache
our $demofile = $CGI->param('file');
our $cleanup = $CGI->param('cleanup');
our $featmat = $CGI->param('featmat');
our $lang = $CGI->param('lang');
$lang = "spanish" if (!$lang);
my $showlang = ucfirst($lang);
our $vbs = $CGI->param('verbose');
our $source = $CGI->param('source');

#########################
# General configuration
#########################

my $minpond = 0.001; # this is an empirically defined parameter 

# labels for the languages
my %tr = (
"spanish" => 'es',
"english"=> 'en',
"french" => 'fr',
);
# respect for the English ortography
my %up = (
"spanish" => 'Spanish',
"english"=> 'English',
"french" => 'French',
);

# the root node
my %tout = (
'spanish'=> 'todo',
'french'=> 'tout',
'english'=> 'anything',
);
# the upper level nodes
my %topont = (
'spanish'=> 'todo|entidad|eventualidad|propiedad|grupo',
'french'=> 'tout|entité|propiété|groupe|éventualité',
'english'=> 'anything|entity|property|group|eventuality',
);

my %store = ( # the names of the folders with the cache
# pending: change them to something more transparent
'spanish'=> 'kindStore',
'french'=> 'wikiFrtore',
'english'=> 'englishStore',
);

my $home = "/usr/lib/cgi-bin/kind/2020/";
# path to this same script

my $webdir = "/home/rogelio/web/"; 
# this is the web directory of the server (at tecling.com)

my $coreont = $home.'coreont_'.$tr{$lang}.".txt";
# this it the coreont (a.k.a. the Top Ontology)

my $features = $home.'features/features2020_'.$tr{$lang}.".txt";
# automatically generated features used for the classification

my $vecfold = $home."vectors-".$lang."/";
# the folder where the co-occurrence vectors are 

my $disk2 = "/media/disk2/"; 
# this is a secondary disk where we are storing most of the data

my $contexts = $disk2.$store{$lang}."/contexts/"; 
# this is a cache where contexts of occurrence of searched terms are stored

my $basetax = "cleantax/taxo_".$lang."2020.txt";
# this is the results of one of the modules which uses dictionaries
# it is stored here because, unlike the rest of them, which work on 
# corpora, this is static.

# here we are just cheching that everything is in place
# If something is not, it will just complain, but it won't stop the execution
foreach my $route ($coreont, $features, $disk2, $contexts, $basetax, $vecfold) {
	print "\n<br>Not found: $route" if (! -e $route); 
}

my $cachedir = $disk2.$store{$lang}."/chains/"; 
# This is the cache. Every time we process a term, we store it there.

my $maxdispers = 2000; # the maximum dispersion, used by the automatic 
		# feature extraction module. It will ignore features
		# that are present in more than those categories.
		# Probably not a good idea.
		# pending: find a better way.

#########################
# Module configuration
#########################

my %umbrales; # these are thresholds, which are different for every module.
my %maxcompmast; # the max number of components per class

#########################
# Co-occurrence selection
# sintagm

$umbrales{sintagm} =  0.02; 
# proportion of contexts where the hypernymy is found

#########################
# Distributional similarity between hyponym-vectors and hypernym-vectors
# paradigm
$umbrales{paradigm} =  13; 
# min. number of components in common
my $dimens = 100; # max number of components per vector
$maxcompmast{paradigm} = 200; # max number of components per class

#########################
# Morphological similarity between co-hyponyms
# morfeo
$maxcompmast{morfeo} = 50;

#########################
# Asymmetric
# asymmetric
$umbrales{asymmetric} = 8; 
# max number of co-occurrent terms that lead to the hypernym


#########################
# Palex
# This is the most resource-intensive module
# and the only one that is language-specific
# see sub palex
# for more in-line documentation
#########################

my (  
# pending: document each of these variables.
$noun,
$middle,
$conj,
$ref, 
@kill,
%examp, 
%kill,
%candidatos, # candidates are stored here
%boosty, # some candidates get a boost
%cohyperonyms,
%tree,
$fondo, 
$data,
@regex,
$art,
);

my $word = '[^ 0-9\t\.,;\(\)\/&%\$=!"\']+'; # Regex for a word
my $kwilema = "\t###\t $word \t###\t"; # words painted by KWiCo
$ref = '\( *([A-Z][a-z]+ *[;&,] *[12][890][0-9][0-9][;&,]?)+ *\)?'; 
# references

if ($lang eq "spanish" ) {
	# pending: this should have an order
	# not alphabetical, but by categories
	@kill = qw (@card@ - , ; : ' “ \" – align prnl a además adj al alguno año aquel así bien bueno cada \@card@ caso clase como con contra cosa cuando cual dar de deber decir del desde determinado diferente donde dos ej ejemplo el él ellos en enlace entre es ese estar este estos etc etc\.? etcétera externo factor familia fenómeno font grande haber hacer hasta importante inglés ir josé juan l la las llegar llevar los lugar m más mayoría mío mismo mucho muy nacional nc necesario ni no nombre nos nosotros np nuevo o otro otros para parecer parte pasar pero plural poder por porque primero que quedar querer quien s santo se seguir segundo ser servicio si sigla sin slash sobre social suyo tal tales también tan tanto te tener típico tipo título tod todo tratar tres un uno url uso variedad varios ver vez nombre name style center article diff class plainlinks text-align year vladj vlfin y ya yo such small sub url a able about across after all almost also am among an and any are as at be because been but by can cannot could dear did do does either type else ever every example important for from get got had has have he her hers him his how however i if in into is it its just least let like likely may me might most must my neither no nor not of off often on only or other our own rather said say says she should since so some than that the their them then there these they this tis to too twas us use wants was we were what when where which while who whom why will with would yet you your yes);
	$noun = 'nc|np|adj|acrnm'; 
	$middle = 'del?|la|el|no|por'; # it could be inside a multiword term
	$art = "las?|el|los";# determiners
	$conj = "y|o";
	# What follows is the cascade of Hearst patterns in Spanish:
	@regex = ($kwilema.'(,|es)? un tipo de '.$word,
	$word.' , como ($art) '.$kwilema,
	$word.',? (que incluyen|incluyendo|inclu[ií]d[oa]s?) (a?('.$art.')? ('.$word.'|'.$kwilema.'))',
	$word.' tales como ('.$art.')? '.$kwilema,
	'('.$word.' |'.$kwilema.')\(( '.$kwilema.'| '.$word.' ),',
	$word.' , entre ellos ('.$kwilema.'|'.$word.')',
	'otros '.$word.' como ('.$art.')? ('.$kwilema.'|'.$word.')',
	'otros ('.$word.' ){1,}\( ('.$kwilema.'|'.$word.')',
	$word.' \( por ejemplo ,? ?('.$art.')? ('.$kwilema.'|'.$word.')',
	'('.$kwilema.'|'.$word.' )\)? ,? y otros '.$word,
	$word.' ,? ?entre (é|e)stos ('.$kwilema.'|'.$word.')',
	'(tipos de '.$word.',? (dentro de|entre) l[oa]s cuales se encuentran? (el|la)? ('.$word.'|'.$kwilema.')(, (el|la)? ('.$word.'|'.$kwilema.'))* ([yeou] (el|la)? ('.$word.'|'.$kwilema.'))+)');

	$data = "
hypo	tal como
hypo	entre el
hypo	entre esto
hypo	por ejemplo
hypo	por~ejemplo
hypo	un tipo de
hyper	y otro
";

} elsif ($lang eq "english" ) {
	@kill = qw (@card@ - , ; : ' “ \" – a adj al año aquel bien bueno cada \@card@ caso clase como con cosa cuando dar de deber decir del desde determinado diferente donde dos ej ejemplo el él ellos en enlace entre es ese estar este estos etc etc\.? etcétera externo factor fenómeno font grande haber hacer importante inglés ir josé juan l la las llegar llevar los lugar m más mayoría mío mismo mucho muy nacional nc necesario ni no nombre nos nosotros np nuevo o otro otros para parecer parte pasar pero plural poder por primero que quedar querer quien s santo se seguir ser servicio si sigla sin slash sobre social suyo tal tales también tan tanto te tener típico tipo título tod todo tratar tres un uno url uso variedad varios ver vez nombre name style center article diff class plainlinks text-align year vladj vlfin y ya yo such small sub url a able about across after all almost also am among an and any are as at be because been but by can cannot could dear did do does either type else ever every example important for from get got had has have he her hers him his how however i if in into is it its just least let like likely may me might most must my neither no nor not of off often on only or other our own rather said say says she should since so some than that the their them then there these they this tis to too twas us use wants was we were what when where which while who whom why will with would yet you your yes first second third fourth fifth include january february march april may june july august september october november december more large each album during many good over against between through high later under);

	# language specific information:
	$noun = 'nns?|np|jj|rbs'; 
	$middle = 'of|the|for'; 
	$art = "the";
	$conj = "and|or";

	@regex = (
	$kwilema.' ?(is|are|was|were) (an?|another|the|any|examples? of) (most|only)? ?'.$word,

#HYPO is a type of HYPER
#HYPO is a HYPER
#HYPO is an HYPER
#HYPO is HYPER
#HYPO is another HYPER
#HYPO is the HYPER
#HYPO is the only HYPER
#HYPO is any HYPER
#HYPO is example of HYPER
	#HYPO is adj sup HYPER
	#HYPO is adj sup most HYPER
#HYPO are HYPER
#HYPO are a HYPER
#HYPO are examples of HYPER
	#HYPO are adj sup HYPER
	#HYPO are adj sup most HYPER
#HYPO was a HYPER
#HYPO were a HYPER

$kwilema.' ?,?a(s|n)? (an?|well as the|(kind|form) of) ?'.$word,

#HYPO as a HYPER
#HYPO as an HYPER
#HYPO as well as the HYPER
#HYPO an HYPER
#HYPO a HYPER
#HYPO a kind of HYPER
#HYPO as HYPER
#HYPO, a kind of HYPER
#HYPO, a form of HYPER

$kwilema.' (another|like|and|or) (the)? ?(many|any|some)? ?(other) '.$word,

#HYPO another HYPER
#HYPO like other HYPER
#HYPO and other HYPER
#HYPO and many other HYPER
	#HYPO and in other HYPER
#HYPO and any other HYPER
#HYPO and some other HYPER
#HYPO or other HYPER
#HYPO or any other HYPER
#HYPO or the many HYPER

$kwilema.' ?'.$word.' types',

#HYPO HYPER types

' (such|examples of) '.$word.' (is|are|as) ?'.$kwilema,

	#Adj sup HYPER is HYPO
	#Adj sup most HYPER are HYPO
#such HYPER as HYPO
#examples of HYPER is HYPO
#examples of HYPER are HYPO

$word.' (among them|except|as|(notab|usual|most|main|principal)ly|like|even|(i|e)\.(e|g)\.|other than|but not|not least|in particular) '.$kwilema,

	#HYPER types HYPO
	#HYPER compared to HYPO
#HYPER among them HYPO
#HYPER except HYPO
#HYPER as HYPO
#HYPER notably HYPO
#HYPER usually HYPO
#HYPER mostly HYPO
#HYPER mainly HYPO
#HYPER principally HYPO
#HYPER like HYPO
#HYPER even HYPO
#HYPER i.e. HYPO
#HYPER e.g. HYPO
#HYPER called HYPO
#HYPER or HYPO
#HYPER other than HYPO
#HYPER but not HYPO
#HYPER not least HYPO
#HYPER in particular HYPO

$word.' (such as|including|(especial|particular)ly|other than|for example) ?(the|a)? ?'.$kwilema,

#HYPER such as HYPO
#HYPER such as the HYPO
#HYPER such as a HYPO
#HYPER including HYPO
#HYPER including the HYPO
#HYPER especially HYPO
#HYPER especially the HYPO
#HYPER particularly the HYPO
#HYPER particularly HYPO
#HYPER other than the HYPO
#HYPER for example HYPO
#HYPER for example the HYPO
' \([^\(\)]+###',

);

	$data = "
hypo	such as
hypo	for instance
hypo	for example
hypo	a kind of
hypo	is a
hyper	an other
";

} elsif ($lang eq "french" ) {
	@kill = qw (@card@ - , ; : ' “ \" – ainsi jean pierre louis paris france français même lui-même nouveau très aussi être dans vers après contre entre sans grand petit pour avoir pouvoir espèce certain avec deux leur autre celui plus pièce comme tout faire premier elle mais catégorie variété série genre type famille style center left right font-size span align small background-color bgcolor chez cest cela janvier février mars avril mai juin juillet août septembre octobre novembre décembre année vous utiliser suivant);
	$noun = 'nom|nam|adj';
	$art = "(au|aucun|aucune|aux|ce|cent|cents|certaines|certains|ces|cet|cette|chaque|cinq|cinquante|de|des|des|deux|différentes|différents|divers|diverses|dix|douze|du|force|huit|huitante|l\'|la|laquelle|le|lequel|les|lesquelles|lesquels|leur|leurs|ma|maint|mainte|maintes|maints|mes|mille|mon|neuf|nonante|nos|notre|nul|nulle|nulles|nuls|octante|onze|plusieurs|quarante|quatorze|quatre|quel|quelle|quelles|quelqu|quelqu|quelque|quelques|quels|quinze|sa|seize|sept|septante|ses|six|soixante|son|ta|tel|telle|telles|tels|tes|ton|tous|tout|toute|toutes|treize|trente|trois|un|une|vingt|vingts|vos|votre|zéro)";
	my $etre = "(es|est|furent|fus|fusse|fussent|fusses|fussiez|fussions|fussé|fut|fûmes|fût|fûtes|sera|serai|seraient|serais|serait|seras|serez|seriez|serions|serons|seront|soient|sois|soit|sommes|sont|soyez|soyons|suis|étaient|étais|était|étant|étiez|étions|été|êtes|être)";
	@regex = (
	$art.' '.$word.' (à savoir|(et plus )?(spécialement|spécifiquement|particulièrement)|par exemple|ex.|en particulier|excepté|à l\'exception de|sauf|tel que|notamment|généralement|et surtout|et principalement|comme|autre que|mais pas|c\'est à dire|en l\'occurrence|principalement|dont) '.$art.' '.$kwilema,
	#DET HYPE <tel> que DET HYPO	!!!
	#DET HYPE à savoir DET HYPO
	#DET HYPE spécialement DET HYPO
	#DET HYPE spécifiquement DET HYPO
	#DET HYPE et plus spécialement DET HYPO
	#DET HYPE et plus spécifiquement DET HYPO
	#DET HYPE par exemple DET HYPO
	#DET HYPE ex. DET HYPOs
	#DET HYPE en particulier DET HYPO
	#DET HYPE particulièrement DET HYPO
	#DET HYPE et plus particulièrement DET HYPO
	#DET HYPE excepté DET HYPO
	#DET HYPE à l'exception de DET HYPO
	#DET HYPE sauf DET HYPO
	#DET HYPE tel que DET HYPO
	#DET HYPE notamment DET HYPO
	#DET HYPE généralement DET HYPO
	#DET HYPE et surtout DET HYPO
	#DET HYPE et principalement DET HYPO
	#DET HYPE comme DET HYPO
	#DET HYPE autre que DET HYPO
	#DET HYPE mais pas DET HYPO
	#DET HYPE c'est à dire DET HYPO
	#DET HYPE à savoir DET HYPO
	#DET HYPE en l'occurrence DET HYPO
	#DET HYPE ou DET HYPO
	#DET HYPE principalement DET HYPO
	#DET HYPE autre que DET HYPO
	#DET HYPE dont DET <exemple> <être> DET HYPO	???
	#DET HYPE <comparer> à+aux DET HYPO	???
	#DET HYPE dont DET HYPO
	$art.' '.$kwilema.' (comme|ainsi que|constitu[^ ]+|'.$etre.'|ressembl[^ ]+) '.$art.' '.$word,

	#DET HYPO <être> DET HYPE	!!!
	#DET HYPO comme DET HYPE
	#DET HYPO ainsi que DET HYPE
	#DET HYPO <constituer> DET HYPE	!!!
	#DET HYPO <être> DET HYPE	!!!
	#DET HYPO qui <ressembler> à\+aux DET HYPE

	$kwilema.'('.$etre.'|,|\() '.$art .' '.$word,

	#DET HYPO DET HYPE

	$art.' '.$kwilema.' ((et|ainsi que|et dans|ou|'.$etre.') )?'.$art.' autre '.$word,

	#DET HYPO DET autre HYPE
	#DET HYPO et DET autre HYPE
	#DET HYPO ainsi que DET autre HYPE
	#DET HYPO et dans DET autre HYPE
	#DET HYPO ou DET autre HYPE
	#DET HYPO <être> DET autre HYPE	!!!

	$art.' '.$kwilema.' (('.$etre.' )?(un type|'.$art.' exemple)|(, )?'.$art.' (sorte|forme)|(, )?(une sorte|une espèce|un genre)) de '.$word,	#5

	#DET HYPO <être> un type de HYPE
	#DET HYPO DET sorte de HYPE
	#DET HYPO <être> DET exemple de HYPE
	#DET HYPO , DET sorte de HYPE
	#DET HYPO , DET forme de HYPE
	#DET HYPO , une sorte de HYPE
	#DET HYPO , une espèce de HYPE
	#DET HYPO , un genre de HYPE
	#DET HYPO une sorte de HYPE
	#DET HYPO une espèce de HYPE
	#DET HYPO un genre de HYPE

	$art.' '.$kwilema.' ((comme|ainsi que) d\'autres|(ou|et) n\'importe quel autre|('.$etre.' ) n\'importe quel|, l\'un de(s| ces)|en tant que) '.$word,

	#DET HYPO comme d'autres HYPE
	#DET HYPO ainsi que d'autres HYPE
	#DET HYPO ou n'importe quel autre HYPE
	#DET HYPO et n'importe quel autre HYPE
	#DET HYPO <être> n'importe quel HYPE	!!!
	#DET HYPO , l'un des HYPE
	#DET HYPO , l'un de ces HYPE
	#DET HYPO <dénommer> HYPE	???
	#DET HYPO <appeler> HYPE	???
	#DET HYPO également <dénommer> HYPE	???
	#DET HYPO également <appeler> HYPE	???
	#DET HYPO ainsi que de nombreux+d'autres HYPE	???
	#DET HYPO <être> DET plus ADJ des HYPE	???
	#DET HYPO en tant que HYPE

	$word.' ou '.$kwilema,

	#HYPE ou HYPO

	$art.' '.$word.' (y compris|appelé|qui '.$etre.' du même type que) '.$kwilema,

	#DET HYPE y compris HYPO
	#DET HYPE appelé HYPO
	#DET HYPE qui <être> du même type que HYPO	!!!

	'(d\'autres|'.$art.' autre) '.$word.' que '.$art.' '.$kwilema,

	#d'autres HYPE que DET HYPO
	#DET autre HYPE que DET HYPO

	'\([^\)]* *'.$kwilema.' *[^\)]*\)',

);

	$data = "";	
}

###############
# Configuration of palex ends here
########################

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

###############
# Configuration of modules (in general) ends here
########################

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

###############
# Configuration of Kwico starts here (Kwico is not a module)
# Kwico (keyword in context)
# is a module of the Jaguar project ( http://www.tecling.com/jaguar )
# which is used here to quickly extract contexts of occurrence of the
# terms from the corpus
# pending: it should be included here as a module instead of being called 
# as a script
###############

my $path = $webdir."/kwico/sourcecode/"; 
# this is where kwico's code is stored
my $max = 5000; 
# max number of contexts extracted per searched term

my %corpora = ( 
# the corpora from where concordances are extracted
'spanish'=> [
	"$webdir/kwico/corpus/eswiki",
	"$disk2/corpus/esten" , # "esTenTen Spain",
	"$disk2/corpus/esamten", # "esTenTen America",
	],
'french'=> [
	"$webdir/kwico/corpus/frwiki",
	"$disk2/corpus/frten",
	],
'english'=> [
	"$webdir/kwico/corpus/enwiki",
	"$disk2/corpus/enten",
	],
);

my $kwico = "/usr/bin/perl $path/kwico.pl"; 
# path to the kwico script

# pending: at some point we tried this to stop verbs pouring in.
# This is because the corpus has a lot of POS-tagging errors, at least
# in Spanish
my %verbmorf = (
'spanish'=> 'r?[eai]rl[oa]s?|[éá]ndo[nl][eoa]s|entamos|eramos|eremos|erimos|aremos|onamos|[aei]rse|[nmcb][eia]mos|[eai]rnos|[ií]amos|rsel[ao]s|[aei]rles|[aei]mos|rear|[vielz]+ar',
'french'=> '', 
'english'=> '',
);

my %flex = (
# Figures of evaluation are estimated on the basis of the results of the automatic contrast against the reference value of evaluation datasets, which indicate the real true/false value of each hypernymy link. This allows us to represent the result of each trial in the standard terms of precision, recall and F1.
# The script that produces the comparison with the gold standard includes what we call a ``flexibility table'' (flextab, for short), to account for synonyms or otherwise acceptable answers from the algorithm. For instance, 
# in Spanish there are two words for \textit{fish}: \textit{pescado} and \textit{pez}. The first is used to
# to refer to fish that has been caught and is intended for eating, while the other one is used the denote the animal in general. In the case of edible fish, sometimes the algorithm proposes the hypernym \textit{pescado}, and consequently the term is classified as a type of food. This is a problem of regular polysemy \cite{Apresjan1974} but, with the flextab, the evaluation script will accept both forms as correct, even when it is only the form \textit{pez} that is attested in the gold standard.
	"animal" => "pez",
	"organismo"  => "pez",
	"familia"  => "pez",
	"pescado" => "pez",
	"carne" => "pez",
	"alimento" => "pez|queso",
	"sistema" => "máquina",
	"aparato" => "máquina",
	"equipo" => "máquina",
	"instrumento" => "máquina",
	"motor" => "máquina",
	"vehículo" => "máquina",
	"barco" => "máquina",
	"artefacto" => "arma|máquina",
	"sustancia" => "substancia",
	"outil" => "machine",
	"instrument" => "machine",
	"artefact" => "machine",
	"aliment" => "poisson|fromage|herbe|légume|céréal|graine|alimentation|plat",
	"viande" => "poisson|porc|saucisse",
	"animal" => "poisson",
	"engine" => "machine",
	"device" => "machine",
	"artifact" => "machine",
	"tool" => "machine",
	"food" => "cheese|fish",
	"sciences" => "science",
	"study" => "science",
	"plante" => "herbe|légume|fruit",
	"boisson" => "alimentation|liqueur",
	"fruit" => "graine",
	"connaissance" => "science",
	"étude" => "science",
	"centre" => "center",
	"human" => "person",
	"plant" => "vegetable",
	"drink" => "beverage",
	"drinks" => "beverage",
	"sauce" => "condiment",
	"noodles" => "noodle",
	"building" => "edifice",
);


#################
# Configuration in general ends here
###############

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#################
# The web interface starts here.
# We print some welcoming html.
###############

my $tit = "Kind - The Taxonomy Project";
our %children;
print $CGI->start_html (
 -title=> $tit,
 -style=>{-src=> ["/style/stylesheet-kind.css", "/style/style-kind-2018.css"], -code=>"body{font-family: 'telluralregular';}" },
 -encoding =>'utf-8',
 -script=>{ type=>'javascript',
        },
);

print "<body>
<br/>
<table width=100% border=0>
<tr><td width=300 align=right><a href='$self?lang=$lang'><img src='/images/kind-logo.png' width=245 height=173></a></td>
<td align=left></td>
</table>
<h2>$tit</h2>";

if (!$source) {
	print "<h4>You are now on the ". $showlang
	." side of the taxonomy</h4>
	<p class='black'> You can also go to the other sides:
	<ul>";
	foreach my $l (sort keys %tr) {
		print "<li> --&gt; <a href='$self?lang=$l'>
		<b>$up{$l}</b></a>" if ($lang ne $l);
	}
	print "</ul>";
}

$self .= "?lang=$lang";
my %datasets = ( 
# we offer some evaluation datasets to the user
'spanish'=> "	<ul>
	<li> + <a href='$self&usecache=1&file=jisys_spanish.txt'>358 single nouns in the categories of weapons, cheese, fish and machines</a>
 	<li>  +<a href='$self&usecache=1&file=1C.spanish.training.gold.csv'>674 terms of various classes (SemEval 2018)</a> (this one is very noisy)
	<li> + <a href='$self&usecache=1&file=1C.spanish.test.gold.txt'>1C.spanish.test.gold.txt</a>
	<li>+ <a href='$self&usecache=1&file=pharma.txt'> 1767 names of drugs and pharmaceutical products</a>
	</ul>
	<br>These others are  single nouns of various categories, not tagged:
	<ul>
	<li>  +<a href='$self&usecache=1&file=874sustantivos.txt'>874 nouns</a>
	<li>  +<a href='$self&usecache=1&file=1314sustantivos.txt'>1314 nouns</a>

	</ul>

	<br>And these are some terms, also not tagged:
	<br>+ <a href='$self&usecache=1&file=psiterms.txt'> 1710 psychiatry terms</a>",


'french'=> "	<ul>
	<li> + <a href='$self&usecache=1&file=jisys_french.csv'>291 single nouns in the categories of weapons, cheese, fish and machines</a>
<li> + <a href='$self&usecache=1&file=environment_eurovoc_fr.taxo'>environment_eurovoc_fr.taxo</a>
<li> + <a href='$self&usecache=1&file=food_fr.taxo'>food_fr.taxo</a>
<li> + <a href='$self&usecache=1&file=food_wordnet_fr.taxo'>food_wordnet_fr.taxo</a>
<li> + <a href='$self&usecache=1&file=science_eurovoc_fr.taxo'>science_eurovoc_fr.taxo</a>
<li> + <a href='$self&usecache=1&file=science_fr.taxo'>science_fr.taxo</a>
<li> + <a href='$self&usecache=1&file=science_wordnet_fr.taxo'>science_wordnet_fr.taxo</a>
	</ul>
	",


'english'=> "	<ul>
	<li> + <a href='$self&usecache=1&file=jisys_english.csv'>273 single nouns in the categories of weapons, cheese, fish and machines</a>
<li> + <a href='$self&usecache=1&file=1A.english.training.gold.txt'>1A.english.training.gold.txt</a>
<li> + <a href='$self&usecache=1&file=2A.medical.training.gold.txt'>2A.medical.training.gold.txt</a>
<li> + <a href='$self&usecache=1&file=2B.music.training.gold.txt'>2B.music.training.gold.txt</a>
<li> + <a href='$self&usecache=1&file=science_en.taxo'>science_en.taxo</a>
<li> + <a href='$self&usecache=1&file=food_en.taxo'>food_en.taxo</a>
<li> + <a href='$self&usecache=1&file=1A.english.test.gold.txt'>1A.english.test.gold.txt</a>
<li> + <a href='$self&usecache=1&file=2A.medical.test.gold.txt'>2A.medical.test.gold.txt</a>
<li> + <a href='$self&usecache=1&file=2B.music.test.gold.txt'>2B.music.test.gold.txt</a>
	</ul>
	",
);

# This is the link for documentation
# We keep in a separate file to facilitate edition.
if ($source) {
	my @doc = &readfile("documentation.htm");
	my $doc = join '', @doc;
	print $doc;
	exit;
}

if (!$input) { 
# if there is no input, we kindly ask for some.
	print "\n<p class='black'>".
	"Enter an arbitrary word or a list of words in <b>$showlang</b> 
	to classify
	(one per line):
	<form action='$self' method='post'>
	<table boerder=1><tr><td valign=top>
	<textarea name='input' cols=15 rows=5></textarea> 
	<input type=hidden name=lang value=$lang>
	<p class='black'><input type=checkbox name=verbose> Verbose 
	<br>(this will make it very verbose, so don't tick it if you are
	<br>puting many input words, otherwise you will get fluded with text).
	<p class='black'><input type=checkbox name=ignorecache> Ignore cache 
	<br>(this will make it not-use the cache, thus it will ignore
	<br> any previous classification of the same noun that it may
	<br> have done in the past.
	<p class='black'><input type=submit value='send'>
	</td>
	
	<td valign=top>
	<font size=2>
	<br>You can also try with these precompiled examples.
	<br>They have already been tagged by humans, so the evaluation 
	can be made automatically.
	$datasets{$lang}
	</font>
	</td></tr></table>";	
}

# and the usual institutional info.
&institutional() and exit if (!$input && !$demofile && !$cleanup && !$featmat);

our %vocab; 
# this is used to store the singular forms of terms. 
# It collects the info from various sources, but mainly the contexts
# of occurrence of analyzed terms
# It is initialized as "our" because it is shared by various modules

our %stop; # a simple stoplist. Also shared by all modules

my (
 @input, # the input list
%input, # the input is turned into a hash to avoid repetitions
 %features, # the automatically generated features (used to classify)
%termfeat, # the features of the input term
%dicco, # stores the hypo-hyper pairs from basetax for the module that
	# uses dictionaries
%rank, # the rank of hypernyms for the classification.
	# it is global but reinitialized in each cycle
 %coreont, # where the coreont is stored
 %hipercoreont,# the nodes of the coreont
%antiloop, # used to avoid cycles in the taxonomy

# the following are used in the automatic feature extraction
%hiperhipo, 
%hipohiper, 
%ctoshipos,
%classfeat,
%featdisp,

 $length, # the length of a taxonomy chain
 $pop, # the last element of a taxonomy chain
%padres,  # the last parents recorded according to the length of the chains
%upward, # the bottom-up path
%not, # used to signal that a category is excluced. It is also renewed
 $stop, # used by the stoplist
$inclasificable, # when a noun cannot be classified
 %hotkeys, # these are the different categories to choose from
@yata, # when an element is already in the coreont, it return the chain
%extras, # used for intermediate hypernyms, such as "pescado azul" 
	# between "caballa" and "pescado"
%chain, # the ascending chains
%lineage, # the elements in an ascending chain
%consider, # meant to be a polysemy detector
	# it will flag a noun as a possible case of polysemy
%contextos, # the contexts of occurrence of an input term
%score, # the final score of each hypernym
%hipotesis, # different hypernymy hypotheses
%cache, # the cache before storing it to disk
);

#########################
# 3.0 morfrules
# see the documentation for details on this module
my (%morfemas, 
%exceptuando, 
%conceptos, 
%repel, 
%killhiper);


#######################
# To switch modules on and off
#######################

#We were interested in 
#measuring the contribution of each individual module to the overall result; therefore, we tested the performance of the algorithm by repeating experiments and switching off one module at a time. 

my %mods = (
'morfrules' => 1, 
'sintagm' => 1, 
'morfeo' => 1, 
'paradigm' => 1, 
'asymmetric' => 1,
'dicco' => 1,
'head' => 1,
'palex' => 1,
'newsynt' => 1,
);

&loadmorfrules() if ($mods{morfrules});

@input = split /\r*\n+/, $input;

if (!$featmat) {
	# featmat is the instruction to extract features automatically
	# from the prunned taxonomy.
	
	# The following operations only have sense if we *have not*
	# been asked to extract the features
	
	# We load dicco's output only after the prunning performed by 
	# the "clean up" module
	# print "\n<br>Loading dicco's output..."  if ($vbs);
	# pending: we have a function to read files
	open( my $fh, '<', $basetax ) or die "\n<br>Error: I cannot read $basetax: $? $!";
	while (my $line = <$fh> ) {
		chomp $line;
		my ($hipo, $hiper) = split /\t/, $line;
		next if (!$hipo || $hipo =~ /[\/\@0-9\* ]+/ || $hipo =~ /^\-/ || $hiper =~ / /); 
		$dicco{$hipo}{$hiper}++;
	}
	close $fh;
	print "\n Ok..." if ($vbs);

	##################
	# Loading the feature matrix
	if (-e $features) {
		print  "\n<br>Loading feature matrix..." if ($vbs);
		open( $fh, '<', $features ) 
			or print "\n<br>I can't read $features: $? $!" 
			and die;
		while (my $line = <$fh> ) {
			chomp $line;
			my @l = split /\t/, $line;
			my $mod = shift @l;
			my $cat = shift @l;
			if (scalar(@l)) {
				foreach my $feat (@l) {
					$feat =~ s/ *~.+//g;
					$features{$mod}{$cat}{$feat}++;
					if ($mod eq "paradigm") {
	 					$vocab{$feat}++ 
					}
				}
			}
		}
		close $fh;
	}
}

# The following variables are used for evaluation
my %eval;
my %goodones; # the number of "good ones" perl category
my %evalterms; # the total vocabulary that is evaluated
		# it is used to populate the dataset with distractors
my ($tp, $fp, $tn, $fn); # the usual true positive, false positive, etc.

# we have an anotated dataset which we use for automatic evaluation
if ($demofile) {
	##################
	print "\n<br>Loading demo file for autoevaluation: [$demofile]";
	##################
	my @eval = &readfile($home.$demofile);
	foreach my $input (@eval) {
		chomp $input;
		$input = lc $input;
		my @campos = split /\t/, $input;
		my $hipo = shift @campos;
		while ($hipo =~ /^[0-9]+$/) {
			$hipo = shift @campos;
		}
		if ($campos[$#campos] =~ /^[01]$/) {
			my $value = pop @campos;
			next if ($value==0);
		}
		$input{$hipo}++;
		foreach my $hiper (@campos) {
			$eval{$hipo}{$hiper}++;
		}
	}
	$goodones{$lang} = scalar keys %eval;
	@input = sort keys %input;
}


##################
# Loading the Coreont
##################

my @coreont = &readfile($coreont);
foreach my $line (@coreont) {
	$line =~ s/ *#.+//g; # ignore comments
	chomp $line;
	next if (!$line);
	if ($line eq "STOPLIST:") {
		$stop++;
		next;
	}
	if ($stop) {
		$stop{$line}++;
		next;
	}
	$line = lc $line;
	my @level = split /\t/, $line; 
	$length = scalar @level;
	$pop = pop @level;
	next if (!$pop);
	# if a chain is one link larger than the previous
	# then it is a hyponym of the last noun.
	# Otherwise, they are children of the last parent recorded
	if ($length) {
		my $hiper = $padres{($length-1)};
		if ($hiper) {
			$vocab{$pop}++;
			$vocab{$hiper}++;
			$upward{$pop}{$hiper}++;
			$coreont{$hiper}{$pop}++;
			$hipercoreont{$hiper}++;
		}
	}
	$hipercoreont{$pop}++; # children can also be future parents
	$padres{$length} = $pop; # we record who is the last parent
}

if ($cleanup) {
	# This is the prunning phase of the taxonomy
	##################
	print  "\n<br>Cleaning up a taxonomy ($cleanup)";
	##################
	my %dirtytax; # the legacy taxonomy (before the prunning)
	my %reality;
	my %exp; # here we store what happened to each pair
	my @file = &readfile($home.$cleanup);
	my $goodones;
	foreach my $input (@file) {
		chomp $input;
		$input = lc $input;
		my @campos = split /\t/, $input;
		my $hipo = shift @campos;
		$hipo = &cleant($hipo);
		next if (!$hipo);
		my $value;
		if ($campos[$#campos] =~ /^[01]$/) {
			$value = pop @campos;
		}
		foreach my $hiper (@campos) {
			next if $hiper =~ /\t[0-9]+\t/;
			$hiper = &cleant($hiper);
			next if (!$hiper);
			if (defined($value)) {
				$reality{$hipo}{$hiper} = $value;
				if ($value) {
					$goodones++;
				}
			} 
			push (@{$dirtytax{$hipo}}, $hiper);
		}
	}
	@input = sort keys %dirtytax;
	my %rank;
	my $c;
	foreach my $hipo (@input) {
		$c++;
		print "\n<br>$c) <b>$hipo</b>: ";
		foreach my $hiper (@{$dirtytax{$hipo}}) {
			my $par = $hipo."_".$hiper;
			my $guess = 0;
			my $head = &head($hipo);
			my $check = $hipo;
			$check = $head if ($head);
			if ($verbmorf{$lang} 
				&& $check =~ /($verbmorf{$lang})$/) {
				# pending: this should be a function
				print "\n<br>This looks like a verb 
				(ends in '-$1'). 
				I will reject it." if ($vbs);
				$exp{$par} = "verbo";
				$guess -= 1000;
			} elsif (&morfgate($check, $hiper)) {
				$guess -= 1000;
				$exp{$par} = "morfgate";
			} else {
				my ($sintag, $hits, $tot) 
					= &sintagm($hipo, $hiper);
				my ($gatnis, $stih, $oto) 
					= &sintagm($hiper, $hipo);
				if ($sintag && $sintag > $gatnis) {
					$guess += (1000 * (1+$sintag)); 
					$exp{$par} = "sintagm";
				}
				my @cache = &cachemira($hipo); 
				foreach my $cach (@cache) {
					if ($hiper eq $cach) {
						$guess += 1000;
						$exp{$par} .= " & cache"; 
						last;
					}
				}
			}
			my $say = "<font color='#cc1111'>false</font>";
			$say = "<font color='#11cc11'>true</font>" 
				if ($guess > 0);
			print "\n<br>---&gt; 
				[$hiper: $say, $exp{$par}, $guess]; ";
			$rank{$par} = $guess;
			delete $contextos{$hipo};
			delete $contextos{$hiper}; # to save memory
		}
	}

	print "\n<br>Saving cleaned up taxonomy..." if ($vbs);
	my $thisfile = '';
	foreach my $par (sort {$rank{$b} <=> $rank{$a} } keys %rank ) {
		last if (!$rank{$par});
		my ($hipo, $hiper) = split /_/, $par;
		$thisfile .= "$hipo\t$hiper\n";
	}

	&savefile($basetax, $thisfile);

	# If the file is annotated, we can evaluate the clean-up
	exit if (!scalar keys %reality);
	print "\n<br>Evaluating cleanup...";
	my ($tp, $fp, $fn, $tn);
	foreach my $par (sort {$rank{$b} <=> $rank{$a} } keys %rank ) {
		my $guess = $rank{$par};
		my ($hipo, $hiper) = split /_/, $par;
		$c++;
		print "\n<hr><hr><br>$c- ($hipo, $hiper): $guess ($exp{$par})";
		if (defined ($reality{$hipo}{$hiper})) {
			print "\n<br>Reality: $reality{$hipo}{$hiper}";
			if ($guess > 0 && $reality{$hipo}{$hiper}) {
				$tp++;
			} elsif ($guess <= 0 && !$reality{$hipo}{$hiper}) {
				$tn++;
			} elsif ($guess > 0 && !$reality{$hipo}{$hiper}) {
				$fp++;
			} elsif ($guess <= 0  && $reality{$hipo}{$hiper}) {
				$fn++;
			}
			print "($tp, $fp, $fn, $tn)";
			my $pre = ($tp + $fp);
			$pre++ if (!$pre);
			$pre = int (100 * ($tp / $pre));
			my $rec = int (100 * ($tp / $goodones));
			my $f1 = 0;
			if ($pre || $rec) {
				$f1 = int (2 * (($pre * $rec) / ($pre + $rec)));
			}
			$rec++ if (!$rec);
			print "\n<p style='color:#111111'>Precision: $pre <br>Recall: $rec <br>F1: $f1";
		}
	}
	exit;
	############
	# End of clean up
	##########
}


if ($featmat) {
	#############
	# Automatic feature extraction
	##########
	print "\n<br>Generating feature matrix from tax $featmat ..." if ($vbs);
	# Important:
	# This can only be done *after* the pruning with cleanup!
	my @file = &readfile($home.$featmat);
	my $c;
	foreach my $line (@file) {
		next if ($line =~ /^(TYPE:|\t|CONFLICT:)/ ); 
		chomp $line;
		my ($hipo, $hiper) = split /\t/, $line;
		next if (!$hipo || $hipo =~ /[\/\@0-9\* ]+/ || $hipo =~ /^\-/); 
		$hiperhipo{$hiper}{$hipo}++;
		$hipohiper{$hipo}{$hiper}++;
		$ctoshipos{$hiper}++;
		if ($mods{morfeo} && length($hipo) > 5 && $hipo =~ /.+(.(.(...)))$/) { 
			foreach my $feat ($1, $2, $3) {
				$classfeat{morfeo}{$hiper}{$feat}++;
				$featdisp{morfeo}{$feat}{$hiper}++;
			}
		}
		$c++;
		&vectorload($hipo,$hiper) if ($mods{paradigm} || $mods{asymmetric} );
		print "\n<br>$c ($hipo,$hiper)" if ($c =~ /00$/);
	}

	print "\n<br>Ok..." if ($vbs);

	#############################
	# Feature weighting starts here
	#############################
	# this is the formula we use to weight the features:	
	# m(x)   1
	# /   .  /
	# t(x)  d(y)
	my %featpond;
	foreach my $mod ('morfeo', 'paradigm') { 
		# only these three modules use the feature weighting matrix
		print "\n<br>Weighting with module $mod..." if ($vbs);
		foreach my $hiper (sort {$ctoshipos{$b} <=> $ctoshipos{$a} } keys %hiperhipo) {
			foreach my $feat(sort{$classfeat{$mod}{$hiper}{$b}<=>$classfeat{$mod}{$hiper}{$a}} 
				keys %{$classfeat{$mod}{$hiper}}){
				next if (!$feat);
				last if ($classfeat{$mod}{$hiper}{$feat} < 3); 
				my $disp = scalar ( keys %{$featdisp{$mod}{$feat}});
				next if (!$disp  || $disp > $maxdispers);
				$featpond{$mod}{$hiper}{$feat} = 
					sprintf "%.5f",(($classfeat{$mod}{$hiper}{$feat}/sqrt($ctoshipos{$hiper}))*(1/$disp));
			}
			my $c;
			foreach my $feat (sort{$featpond{$mod}{$hiper}{$b}<=>$featpond{$mod}{$hiper}{$a}} 
					keys%{$featpond{$mod}{$hiper}}){
				delete $featpond{$mod}{$hiper}{$feat} and next 
					if ($c > $maxcompmast{$mod} || $featpond{$mod}{$hiper}{$feat} < $minpond);
				$c++;
			}
		}
	}

	print "\n<br>Saving datamatrix ($features)..." if ($vbs);
	my $thisfile;
	foreach my $mod (sort keys %featpond) {
		foreach my $hiper (sort keys %{$featpond{$mod}}) {
			my $thisentry = "\n".$mod."\t".$hiper;
			my $c;
			foreach my $feat (sort keys %{$featpond{$mod}{$hiper}}) {
				next if ($feat eq $hiper);
				$c++;
				$thisentry .= "\t$feat ~ ".$featpond{$mod}{$hiper}{$feat};
			}
			if ($c) {
				$thisfile .= $thisentry;
			}
		}
	}
	my $saved = &savefile($features, $thisfile);
	if ($saved) {
		print "\n<br>Ok, data generated at: $saved";
	} else {
		print "\n<br>Something went wrong and I could not save the data! [$? $!]";
	}
	exit;
}



############################
## THE ACTUAL CLASSIFIER STARTS HERE
############################

if ($ignorecache && $vbs) {
	print "\n<br>(Ignoring cache)";
}


&procesalista(@input); # we process the input list

sub procesalista {
   my @input = @_; 
   my $num;
   foreach my $in (@input) {
	$in = &cleant($in);
	next if (!$in);
	next if ($antiloop{$in}); # to avoid infinite cycles
	$antiloop{$in}++;
	undef %chain;
	undef %consider;
	undef %contextos;
	undef %score;
	undef %lineage;
	undef %cache;
	undef %extras;
	undef %hipotesis;
	$num++;
	print "\n\n<hr><hr><hr><p style='color:#111111'><b>$num) $in</b> ";
	# we first inspect if the input term is the corent
	# note the "isfirst" param to the isincore sub
	my $isit = &isincore($in, "isfirst");
	if ($isit) { 
		print "\n<br>The term $in is in the coreont." if ($vbs);
		# and the proces for this term ends here
		# This is a problem because it should process them
		# anyway, otherwise it will never detect polysemy
		# in terms that are in the coreont.
		# Something similar happens with the cache.
		# pending: find a better way.
		print &flushtoponto($in);
		&evaluate($in, $isit,"isincorefirst");
		next;
	}

	my $head = &head($in); # we obtain the syntactic head
				# if it is a multiword unit
	my $check = $in;
	$check = $head if ($head);
	if ($check =~ /($verbmorf{spanish})$/) {
		# Sometimes we get verbs and we have to reject them.
		# pending: we have a function for this
		print "\n<br>This looks like a verb (it ends in '-$1').
			I will skip it."  ;
		next;
	}

	if ($usecache) {
		# This param tells the algorithm to use the cache or not 
		my @cache = &cachemira($in);
		if ($cache[0]) {
			@yata = @cache;
			print &flushtoponto($in); 
			&evaluate($in, $cache[1],"isincache");
			next;
			# As already said in case of the coreont check
			# it should process the terms anyway because
			# it could find new meanings in the case of
			# polysemous terms or semantic neology.
			# pending: find a better way.
		}
	}

	my $paradic = $in;
	$paradic = $head if ($head);
	# it looks upon the basetaxonomy, which only has single terms
	# pero también me fijo en qué me ofrece esa taxo
	# It will only trust the output if the offered hypernym 
	# is in the coreont
	my @dicco = keys %{$dicco{$paradic}};
	if (scalar(@dicco)) {
		print "\n<br> $in ($paradic): it's in the basetaxonomy."  if ($vbs);
		foreach my $hipcand (@dicco) {
			next if ($stop{$hipcand});
			my $suma = 10;
			if (&isincore($hipcand)) {
				print "\n<br>And the hyper $hipcand is in the coreont." if ($vbs);
				$suma += 100;
				# if there is only one hyperny
				# then it is more valuable
				if (scalar(@dicco) == 1) {
					print "\n<br>It's the only candidate." if ($vbs);
					$suma += 1000;
				}
			}
			$hipotesis{$hipcand}++;
			$score{$hipcand} += $suma;
		}
	}

	my $hiper = &classmanager($in);
	if ($hiper && $hiper ne $tout{$lang}) { 
		print  "\n<br>First hypothesis: $hiper ($score{$hiper})" if ($vbs);
		$hipotesis{$hiper}++;
		# the output of classmanager will always be a member of the coreont
	}
	my $orig = $in;
	if ($in =~ /[^i]s/) {
		my $lema = &noplur($in);
		if ($lema) {
			print "\n\n<br>Looks like a plural. I will attempt to lemmatize: <i>$lema</i>" if ($vbs);
			$in = $lema;
			if ($hipotesis{$in}) {
				print "\n<br>Lemma ($in) confirms hypthesis..." if ($vbs);
				$score{$in} += 1000;
			}
			$hiper = &classmanager($in); # we go again
		 	if ($hiper  && $hiper ne $tout{$lang}) {
				print "\n<br>Analysis of lemma produces: $lema ~ $hiper ($score{$hiper})" if ($vbs);
				if ($hipotesis{$hiper}) {
					print     "\n<br>Analysis of lemma ($in) confirms hypthesis ($hiper)..." if ($vbs);
					$score{$hiper}  += 1000;
				}
			}
		}
	}

	if ($head) {
		# pending: here too it should check if the term is in the taxo before it goes on
		if ($hipotesis{$head}) {
			print "\n<br>Head ($head) confirms hypthesis..." if ($vbs);
			$score{$head}  += 1000;
		} elsif (&isincore($head)) { # we check if the head is in the coreont
			print "\n<br>Head ($head) is in the coreont." if ($vbs);
			$score{$head}  += 1000;
		} else { # &classmanager is called again 
			# this time with the head
			$hiper = &classmanager($head);
			if ($hiper && $hiper ne $tout{$lang}) {
				print  "\n<br>Analysis of head results in: $hiper ($score{$hiper})" if ($vbs);
				if ($hipotesis{$hiper}) {
					print "\n<br>Analysis of head ($head) confirms hypthesis ($hiper)..." if ($vbs);
					$score{$hiper}  += 1000;
				} else {
					$hipotesis{$hiper}++;
				}
			}
		}
	}
	# The top-down process ends here,
	# the bottom-up process begins.

	print "\n<br>Analyzing contexts of occurrence..." if ($vbs);
	my ($best, $scoreh) =  &palex($orig);
	# pending: this module should be called palex

	if ($stop{$best} ) {
		undef $best;
	}

	if ($best) {
		$best =~ s/^[^0-9a-záéíóúüñ]+//g; 
		# I have a problem with a filler character
		# and I am trying to get rid of it.
		# pending: find a better way...
	
		my $lema = &noplur($best); 
		$best = $lema if ($lema);
		print "\n<br>Direct hypernym detected: <i>[$best]</i> ($scoreh). " if ($vbs);
		$score{$best} += ($scoreh*2); 
		if ($lineage{$best} ) {
			print "\n<br>$best is confirmed." if ($vbs);
			$score{$best} += 1000; 
		} else {
			my $head = &head($best);
			if (&isincore($best)) {
				print "\n<br>$best is in the coreont. " if ($vbs);
				$score{$best}  += 1000;
			} elsif ($head && &isincore($head)) {
				print   "\n<br>The head [$head] from [$best] is in the coreont." if ($vbs);
				$score{$head} += $score{$best} + 1000;
				$extras{$head} = $best; 
				# this would be like
				# "pescado azul" between "caballa" 
				# and "pescado"
			} 
		}

		if ($best ne $hiper) {
			print "\n<br>Searching now the hypernym of $best" if ($vbs);
			# it checks if the head is in the coreont
			my $head = &head($best);
			if (&isincore($best)) {
				print "\n<br>Hypernym ($best) is in the coreont." if ($vbs);
				$score{$best} += 1000; 
			} 
			if ($head && $head ne $hiper) {
				if (&isincore($head)) {
					print "\n<br>Head [$head] is in the coreont." if ($vbs);
					$score{$head} += 1000; 
				} else {
					$hiper = &classmanager($head); 
					if ($hiper && $hiper !~ $topont{$lang}) {
						print "\n<br>Score $hiper: $score{$hiper}" if ($vbs);
						if ($hipotesis{$hiper}) {
							print "\n<br>Analysis of head [$head] from '$best' confirms hyper [$hiper]" if ($vbs);
							$score{$hiper} += 1000; 
						} else {
							$hipotesis{$hiper}++;
						}
					}
				}
			}

			# We call palex again
			print "\n<br>Analaysing contexts of $best..." if ($vbs);
			my $firsth = $best;
			($best,$scoreh) =  &palex($best);

			if ($best) {
				print "\n<br>The hypernym of $firsth seems to be $best ($scoreh)." if ($vbs);
				$score{$best} += ($scoreh*2); 
				if (&isincore($best)) {
					print "\n<br>The hypernym ($best) is in the coreont." if ($vbs);
					$score{$best} += 1000; 
				} 

				$head = &head($best);
				if ($head && $head ne $hiper && $firsth ne $head && &isincore($head)) {
					print "\n<br>The head [$head] of $best is in the coreont." if ($vbs);
					$score{$head} += 1000; 
				}
			}
		}
	}
	my $chosen;
	my @rank = sort { $score{$b} <=> $score{$a} } keys %score;
	my $frec = $termfeat{paradigm}{$orig}{$orig};
	my $pts; # the point of each candidate
	print   "\n<br><br>Weighting candidates..." if ($vbs);
	while ( my $candidato = shift @rank) {
		next if ($stop{$candidato} || $candidato eq $chosen || $candidato eq $orig);
		next if (&morfgate($orig, $candidato));
		print   "\n<br><br>$candidato ~ $score{$candidato}" if ($vbs);		
		$chosen = $candidato and next if (!$chosen);
		if ($candidato && $candidato ne $orig && $score{$candidato} > 10) {
			my ($sintag, $hits, $tot) = &sintagm($orig, $candidato);
			my ($gatnis, $stih, $oto) = &sintagm($candidato, $orig);
			 print "\n<br><br>sintagm($orig, $candidato)=($sintag, $hits, $tot)<br>sintagm($candidato, $orig)=($gatnis, $stih, $oto)" if ($vbs);
			if ($sintag && $sintag > $gatnis  && $sintag > $pts) {
				$score{$candidato} += (1000 * (1+$sintag)); 
				print "\n (I like this one: $sintag) = $score{$candidato}. " if ($vbs);
				if ($score{$candidato} > $score{$chosen}) {
					print "\nSo $candidato beats $chosen ($score{$candidato} > $score{$chosen})." if ($vbs);
					$chosen = $candidato;
					$pts = $sintag;
				} else {
					print "\nBut it doesn't beat $chosen ($score{$chosen})" if ($vbs);
				}
			} elsif ($score{$candidato} > 15) { 
				# pending: params should not be here
				my $check = &tothetop($candidato, $chosen);
			 	if ($check) {
					print   "\n<br><br>$candidato takes back to $check" if ($vbs);
					$chosen = $candidato;
				}
			} 
		}
	}

	if (!$chosen) {
		print "\n<br>No results for $orig.";
		next;
	}

	my $cadena = &flushtoponto($orig, $chosen, $extras{$chosen});
	if ($cadena) {
		print "\n<br><br>$cadena";
		if (scalar (@rank)) {
			my $consider = '';
			foreach my $candidato (@rank) {
				next if ($lineage{$candidato});
				$consider .=  "\n<br>$candidato: $score{$candidato}"; 
			}
			print "\n<br><font size=2>But consider these options:</font> $consider" if ($consider);
		}
	}
	# Sometimes it cannot form a full taxonomy chain but the selected
	# hypernym is still valid
	# So it tries to do something about it
	if (!$cadena && $score{$chosen} >= 15) {
		print "\n<br>I don't have a full chain but I like the hypernym candidate anyway ($chosen ~ $score{$chosen}). " if ($vbs);
		my @cache = &cachemira($chosen);
		if ($cache[0]) {
			@yata = @cache;
			print &flushtoponto($chosen);
		} else {
			 print "\n<br>Not in cache. Creating new entry..." if ($vbs);
			&procesalista($chosen); 
			# this new call will make a new entry in the cache
			# pending: this is unnecessarily complicated
			# find a simpler solution
			print "\n<br>Ok, reading cache for $chosen..." if ($vbs);
			my @cache = &cachemira($chosen);
			if ($cache[0]) {
				@yata = @cache;
				print &flushtoponto($chosen); 				} else {
				print "\n<br>I could not add the entry to the cache!" if ($vbs);
			}
		}
	}

	if ($chosen ne "todo" && $score{$chosen} < 15 || $chosen =~ /(Empate entre)/ ) {
		print "\n<br>I am not sure of this result";
		print "($chosen ~ $score{$chosen})" unless ($1);
		print " I will ignore it!";
		next;
	}

	my $cachefile = $cachedir.$orig;

	if (-e $cachefile) { 
		# if the cache exists, it checks if it is the same chain
		# if it is different, it will save both
		print "\n<br>Comparing with existing cache..." if ($vbs);
		my @cache = &cachemira($in);
		my $cts;
		my $level;
		my $cache = '';
		foreach my $link (reverse @cache) { 
			# reverse because if goes top-down
			$level++;
			$cache .= "$level) $link --&gt;";
			if ($lineage{$link}) {
				$cts++;
			} 
			last if $level == 5;
		}
		if ($cts < $level) {
			# A difference if found among the first 5 links
			# so they are different chains
			print "\n<br>This version is different from the one in the cache: ".$cache if ($vbs);
			# pending: uncomment this and make it effective
			# $cache{$orig} = "\n\t".$in."\t".$frec.$cache{$orig};
			# open(OUT, ">>".$cachefile) 
			# or print "\nI can't create $cachefile: $!" and die;
			# print OUT $cache{$orig};
			# close OUT;
		}
	} elsif ($cache{$orig}) { 
		# pending:
		# this is repeated. Create a subroutine for it
		$cache{$orig} = "\n\t".$in."\t".$frec.$cache{$orig};
		open(OUT, ">".$cachefile) 
			or print "\n<br>I can't create $cachefile: $!" 
			and die;
		print OUT $cache{$orig};
		close OUT;
	} else {
		print "\n<br>I can't save cache for $orig! (!$cache{$orig} || -e $cachefile)";
	}

	&evaluate($orig, $chosen, "ischosen");
   } # end loop @input
}

###########
# Module subroutines start here 
# pending: a short index of all functions with an explanation

sub morfgate {
	# There were some cases in which morphology was making the
	# classification impossible, and we were forced to apply this
	# new filter.
	# It is only relevant for Spanish. The rest of the languages
	# do not present the same problem.
	my ($hipo, $hiper) = @_;
	my $check = '';
	foreach my $case ('proceso', 'propiedad', 'enfermedad', 'doctrina') {	
		$check .= '|' if ($check);
		$check .= $morfemas{$lang}{$case};
	}
	if ($hiper =~ /^(entidad|persona|animal|animado|máquina|pez|queso|planta)/
			# pending: this should be read from the coreont
			 && $hipo =~ /...($check)$/) {
		# $hipo cannot be a hyponym of $hiper with this morphology
		return "kill";
	}
}


sub evaluate {
	# We use this subroutine for the automatic evaluation
	# with datasets (goldstandards)
	# We go through the generated data collating them with
	# the goldstandard
	my ($orig,$hiper,$caller) = @_;
	return if (!$hiper || !$eval{$orig});
	my $agree;
	my $coma;
	foreach my $myhiper (sort { $lineage{$b} <=> $lineage{$a} } keys %lineage) {
		my @sp = split /\|/, $flex{$myhiper};
		undef @sp if (scalar (@sp) == 1);
		my $head = &head($myhiper);
		if ($eval{$orig}{$myhiper}
			|| $eval{$orig}{$flex{$myhiper}}
			||($sp[0]&&$eval{$orig}{$sp[0]})
			||($sp[1]&&$eval{$orig}{$sp[1]})
			|| ( $head && $eval{$orig}{$head} )
			|| ( $head && $eval{$orig}{$flex{$head}} )
			) {
			$agree++;
		} 
	}
	print "\n<br><br><font size=2>Gold-standard says $orig is a kind of:</font><br>";
	undef $coma;
	foreach my $goldhiper (sort keys %{$eval{$orig}}) {
		my $color = 111111;
		print $coma;
		$coma = ", ";
		my $head = &head($goldhiper);
		if ($head && $lineage{$head}) {
			$agree++;
		}
		print "\n<font color='#$color'>$goldhiper</font>";
	}
	if ($agree) {
		# there is agreement with the goldstandard
		$tp++; 
		$agree = "<font color='#11cc11'>Agreement</font> ";
	} else {
		$fp++; # goldstandard says no but algorithm says yes
		$agree = "<font color='#cc1111'>Disagreement</font>";
	}
	print "\n<p style='color:#111111'>$agree";
	my $pre = ($tp + $fp);
	$pre++ if (!$pre);
	$pre = int (100 * ($tp / $pre));
	my $rec = int (100 * ($tp / $goodones{$lang}));
	my $f1 = 0;
	if ($pre || $rec) {
		$f1 = int (2 * (($pre * $rec) / ($pre + $rec)));
	}
	$rec++ if (!$rec);
	print "\n<p style='color:#111111'>Precision: $pre <br>Recall: $rec <br>F1: $f1";
}

sub classmanager {
	# The purpose of this function is to control the calls to 
	# classify, which is the function that performs the top-down
	# classification
	# classmanager is particularly useful in the case of multiword
	# terms because first it searches for the full term and later
	# for its syntactic head
	my $hipo = $_[0];
	print "\n<br>Classmanager dispatched with [$hipo]." if ($vbs);
	my $largo;
	my $llega; # the maximum length of the head
	my $hiper = $tout{$lang};
	my $abort;
	foreach my $try (1 .. 8) { 
		# pending: these numbers should be params
		my ($back, $score) = &classify($hipo, $hiper);
		last if ($try > 5 && $score < 40 ); 
		# pending: these numbers too should be params
		if (!$back && $try == 1) {
			last;
		} elsif ($back eq "INCLASIFICABLE" || $back =~ /Empate entre/ ) {
			print "\n<p style='color:#111111'>$back" if ($try == 1);
			last;
		} elsif ($back eq "TOPONTOLOGY") { 
			# this means that it is in the coreont
			# but it should have checked that beforehand
			# pending: do it that way
			$score{$hipo} += 10000; 
			# pending: a better administration of these rewards
			last;
		} elsif ($back) {
			$score{$back} += $score;
			print "\n<br>Classmanager produces: $back +$score = $score{$back}" if ($vbs);	
			$hiper = $back;
			last if ($back eq "persona" || $back eq "sistema");
			# pending: avoid these language dependent patches!
		}
	}
	print "\n<br>Classmanager returns [$hiper]."  if ($vbs);
	return $hiper;
}

sub classify {
	# This subroutine receives an input noun and a list of possible
	# hypernyms (that we call the 'hotkeys')
	# It produces a ranking of the available options and 
	# keeps the best available
	# pending: check the competing classes to see if there is 
	# polysemy
	my ($in, $position) = @_;
	undef %rank;
	undef $inclasificable;
	undef %not;
	undef %hotkeys;
	my @cl = sort keys %{$coreont{$position}};
	print  "\n\n<p style='color:#111111'>
	Classifying '$in' among the following categories: <b>".(join ', ', @cl)."</b>" if ($vbs);
	foreach my $l (@cl) {
		$hotkeys{$l}++;
	}

	if (!$termfeat{paradigm}{$in}) {
		print  "\n<br>Loading vectors from $in..." if ($vbs);
		&vectorload($in);
		if ($inclasificable) {
			print "\n\t<p style='color:#111111'>There are no vectors for this noun!" if ($vbs);
		}
	}
	# These modules will only operate if they are active on %mods
	&asymmetric($in); 

	&newsynt($in); 

	&morfrules($in); 

	# Now we check if the noun is in the coreont.
	# The reason for doing it so late is to give the others
	# the opportunity to have a say on the matter, 
	# because someday, we will be detecting polysemy too.
	if (&isincore($in)) {
		print "\n<br>This noun is the coreont." if ($vbs);
		return ("TOPONTOLOGY", 100000);
		# pending: replace this message with "coreont"
	}

	&morfeoload($in); # we extract morphological features of the
			# analyzed term
	my $depth = 5; # this params tells the algorithm how deep it should
			# go to get the features
			# pending: as a params, it should not be here.
	foreach my $class (@cl) {
		# Now we will test, one by one, each of the classes
		# (the hotkeys)
		&absorv($in, $class, $class, $depth);
		&newsynt($in, $class); 
	}
	if (scalar(keys %not) == scalar(keys %rank)) {
		# this would mean that all classes have been rejected.
		undef %not;
	}
	my ($first, $winner);
	foreach my $class (sort {$rank{$b} <=> $rank{$a}} keys %rank) {
		next if ($not{$class});
		if ($stop{$class}) {
			# this class is rejected
			next;
		}
		print  "\n<p style='color:#111111'>$class\t$rank{$class}" if ($vbs);
		if (!$first) {
			$first = $rank{$class};
			$winner = $class;
		} elsif ($first == $rank{$class}) {
			if ($first > 10) {
				# pending: change the message 
				return "</b>Empate entre <b> <i>$winner</i></b> y <b><i>$class</i></b> ($first)...";
			} else {
				# pending: change this message too
				return "INCLASIFICABLE"; 
			}
		}
	}
	if ($winner && $rank{$winner}) { 
		return ($winner, $rank{$winner});
	} else {
		# pending: change this message too
		return "INCLASIFICABLE"; 
	}
}

sub tothetop {
	my ($hipo, $hiper, $dowhat, $caller) = @_;
	print "\n<br><br>tothetop dispatched with ($hipo, $hiper, $dowhat, $caller)" if ($vbs);
	foreach my $y (sort keys %{$upward{$hipo}}) {
		if ($dowhat eq "bring") {
			push (@yata, $y);
		} elsif ($hiper ne "__filler__" && $y eq $hiper) { 
			# "__filler__" is a code to express 
			# that the pos is empty
			return $y;
		} elsif ($dowhat eq "hotkeys" && $hotkeys{$y}) {
			return $y;
		}
		return &tothetop($y, $hiper, $dowhat, 'self') if ($y ne $hiper);
	}
}

sub flushtoponto {
	# When we find a hypernym that is the taxonomy,
	# then we extract its full hypernymy chain
	my ($hipo, $hiper, $extra) = @_; 
	# We suppose that $hipo is the hyponym of $hiper
	# but it could be the case that there is some
	# intermediate link $extra
	my ($largo, $llega);
	print   "\n<br>flushtoponto dispatched with ($hipo, $hiper, $extra)"  if ($vbs);
	if (!scalar(@yata) ) {
		my $send;
		if (&isincore($hipo)) {
			$send = $hipo;
		} elsif ($hiper && &isincore($hiper)) {
			$send = $hiper;
		} elsif ($extra && &isincore($extra)) {
			$send = $extra;
		}
		&tothetop($send, $tout{$lang}, "bring", "flushtoponto");
	} 
	if ($yata[0] !~ /^($topont{$lang})$/ && $yata[$#yata] =~ /^($topont{$lang})$/ ) {
		@yata = reverse @yata;
	} elsif ($yata[0] !~ /$topont{$lang}/ && $yata[$#yata] !~ /$topont{$lang}/ ) {
		return;
	} 
	my $head = &head($hipo);
	push (@yata, $hiper) if ($hiper);
	push (@yata,$extra) if ($extra && $extra ne $hipo && $head ne $extra && $extra ne $yata[$#yata] && $extra ne $yata[($#yata-1)]);
	push (@yata, $head) if ($head && $head ne $hipo && $head ne $extra && $head ne $yata[$#yata]  && $head ne $yata[($#yata-1)]);
	push (@yata, $hipo) unless (!$hipo || $hipo eq "TOPONTOLOGY" || $yata[$#yata] eq $hipo && $hipo ne $yata[($#yata-1)]);
	my $ret = '';
	foreach my $level (@yata) {
		next if (!$level || $level eq $tout{$lang});
		$llega++;
		$largo .= "--";
		$ret .= "\n<p style='color:#111111'>$llega. $largo &gt; <b>$level</b>";
		$cache{$hipo} = "\t".$level.$cache{$hipo};
		$lineage{$level} = $llega;
	}
	undef @yata;
	return $ret;
}

sub morfeoload {
	my $in = $_[0];
	if ($in =~ /.+(.(.(...)))$/) { 
		# $in needs to have more than 5 letters
		$termfeat{morfeo}{$in}{$1}++;
		$termfeat{morfeo}{$in}{$2}++;
		$termfeat{morfeo}{$in}{$3}++;
	}
}

sub absorv {
	# This one is meant to "absorv" the features of the hyponyms
	# e.g., the hypernym "animal" will have the features of its members
	my ($in,$cat, $class, $depth) = @_;
	my $ret;
	foreach my $subclass (sort keys %{$coreont{$cat}}) {
		$rank{$class} += &feam($in, $subclass, $class);
		if ($depth) {
			$depth--;
			&absorv($in,$subclass, $class, $depth);
		}
	}
}

sub feam {
	# it detects feature matches
	my ($in,$cat, $class) = @_;
	my $ret;
	foreach my $mod ('morfeo', 'paradigm') { 
		my $rank; 
		next if (!$mods{$mod});
		foreach my $f (sort { $termfeat{$mod}{$in}{$b}<=>$termfeat{$mod}{$in}{$a}} keys %{$termfeat{$mod}{$in}}){
			# print "\n<br>$mod searching feature [$f] of [$in] within [$cat, $class] (rank: $rank)" if ($vbs);
			$rank++;
			last if ($rank > 20 
				|| $termfeat{$mod}{$in}{$f} < 4); 
			# pending: these are params and should not be here
			if ($features{$mod}{$cat}{$f}) {
				print "\n<br>$mod detects feature [$f] in category [$cat]" if ($vbs);
				$ret++;
			}
		}
	}
	return $ret;
}

sub newsynt {
	return if (!$mods{newsynt});
	print  "\n<br>newsynt is called..." if ($vbs);
	my $in = $_[0];
	my $rank;
	foreach my $cooc (sort{$termfeat{paradigm}{$in}{$b}<=>$termfeat{paradigm}{$in}{$a}} keys %{$termfeat{paradigm}{$in}}){
		next if (!$cooc || $cooc eq $in
			|| $cooc =~ /^(grupo|grupe|grup)$/ ); 
		# pending: this is too language dependent
		# Instead of this, we should have something like %tout 
		$rank++;
		last if ($rank > 15 
			|| $termfeat{paradigm}{$in}{$cooc} < 4); 
		# pending: these should be params
		print "\n<br>newsynt studing: $rank) $in $cooc ~ ($termfeat{paradigm}{$in}{$cooc}) " if ($vbs);
		if ($hotkeys{$cooc}) {
			print "\n<h2>newsynt says: $in $cooc ~ ($termfeat{paradigm}{$in}{$cooc}) $rank</h2>" if ($vbs);
			my $suma = 20 - $rank;
			$rank{$cooc} += $suma;
			$consider{$cooc}++;
			$rank{$cooc} += 5 if ($cooc eq "máquina"); 
			# pending: delete this line (máquina)
			last;
		} elsif ($rank < 10 && &isincore($cooc)) {
			print "\n<br>Calling tothetop..." if ($vbs);
			my $found = &tothetop($cooc, "__filler__", "hotkeys", "newsynt");
			print "\n<br>Back from tothetop..." if ($vbs);
			if ($found ) {
				print "\n<br>newsynt says $cooc takes back to ($found)." if ($vbs);
				my $suma = 20 - $rank; 
				$rank{$found} += $suma;
				$consider{$found}++;
			} else {
				print "\n<br>tothetop did not produce a result." if ($vbs);
			}
			next;
		}
		print "\n<br>newsynt is finished..." if ($vbs);
	}
}


sub vectorload {
	my ($in,$hiper) = @_;
	return if (length($in) < 3);
	my $file = $vecfold.$in; 
	if (! -e $file ) {
		print "\n<p style='color:#111111'> File not found: [$file]. <br>Calling vectorate..." if ($vbs);
		&vectorate($in);
	}
	return 	if (! -e $file ) ;
	open( my $fh, '<', $file ) or print "\n<br> I can't read [$file] because: $? $!" and return;
	my $c;
	while ( my $line = <$fh>) {
		$c++;
		last if ($c > $dimens);
		chomp $line;
		my @t = split /\t/, $line;
		next if ($t[1] eq "V");
		# Here is where I could discriminate by pos
		# and ignore verbs, for example.
		#tacho   N       5151
		#ser     V       1407
		#basura  N       1012
		$termfeat{paradigm}{$in}{$t[0]} = $t[2];
		$classfeat{paradigm}{$hiper}{$t[0]}++;
		$featdisp{paradigm}{$t[0]}{$hiper}++;
	}
	close $fh;
	print "\n<br>Ok!" if ($vbs);
	# I do not return anything because they are globals.
}

sub loadmorfrules {
	# In this sub I lead the morphological rules.

	# These are morphemes associated with semantic categories.
	# Most of them where extracted using morfeo.
	# pending: many more morphemes and rules could be added with
	# the same method.

	# pending: these rules should be centralized somewhere else, 
	# like in the coreont.

	# ~~~~~~~~~~ Spanish rules ~~~~~~~~~~~~~~~~~

	%{$morfemas{spanish}} = (
	'proceso' => 'miento|[sc]ión|ada|azgo|esis|ría|aje|azo|[zndtc]eo|ande',
	'propiedad' => 'dad|idez|encia|nía|anza|ud',
	'adverbio' => 'mente',
	'mueble' => 'teca', 
	'verbo' => 'arse|erse|ear',# pending: delete this because 
				# it is not used anymore
	'ciencia' => 'tica|logía|tría|cultura',
	'técnica' => 'cromía|grama|etría|grafía|tecnia',
	'tratamiento' => 'terapia|tratamiento',
	'cirugía' => 'cirugía|ctomía',
	'instrumento' => 'scopio|metro|fono',
	'fobia' => 'fobia',
	'doctrina' => 'ismo',
	'persona' => 'iente|ópata|ista|[íó]mano|logo',
	'enfermedad' => 'osis|asis|itis|[tl]emia|asmia|patía|manía|n[oe]sia|olia|oma',
	'sustancia' => '[pltn]am|[srnzli]ol|[rdnx]il|[pnmri]ato|ium|sin|nasa|tin|rix|lox|[rt]yl|[rn]ex|dine|feno|trim|ítico',
	'máquina' => 'ora?'
	);

	%{$exceptuando{spanish}} = (
	# These are exceptions to the rules:
	'mueble' => 'karateca|aguacateca|tlaxcalteca|manteca', 
	'doctrina' => 'organismo|microorganismo|abismo',
	'propiedad' => 'actividad|localidad|mudanza|danza|artesanía|ciudad',
	'técnica' => 'malhetría|anagrama',
	'persona' => 'amatista|simiente|pista|inconveniente|aguardiente|estupefaciente|excipiente|reluciente|expediente',
	'proceso' => 'pimiento|encrucijada|establecimiento|estacionamiento|fortificación|habitación|inmediación|instalación|morada|ubicación|yacimiento|edificación|afección|drogadicción|información',
	'ciencia' => 'anfibología',
	'enfermedad' => 'apoptosis',
	'sustancia' => 'fármaco|medicamento',
	'máquina' => 'arma'
	);

	%{$conceptos{spanish}} = (
	# pending: this should be read from the coreont
	'entidad' => 'persona|hombre|mujer|animal|sustancia|objeto|planta',
	'proceso' => 'eventualidad|acción|proceso|acto|estado|actividad|golpe|hecho|guerra|comida|sentimiento|desarrollo|cambio|delito|operación|daño|dibujo|dirección|acuerdo|formación|procedimiento|división|pago|alteración|impresión|violencia|ceremonia|defecto|enfermedad|trastorno|producción|fabricación|aumento|reacción|generación',
	'propiedad' => 'propiedad|cualidad|calidad|estado|falta|carácter|condición|sentimiento|capacidad|dificultad|voluntad|circunstancia|comportamiento|defecto|tristeza',
	'ciencia' => 'estudio|ciencia|conocimiento|medicina|filosofía',
	'técnica' => 'arte|norma|técnica|sistema|diagnóstico|estudio|método',
	'tratamiento' => 'terapia|tratamiento',
	'instrumento' => 'instrumento|aparato',
	'fobia' => 'fobia|miedo|temor',
	'doctrina' => 'arte|doctrina|ideología|ciencia|estilo|estudio|teoría|movimiento|sistema|enfermedad|palabra|estado|actividad|actitud|tendencia|deporte|ejercicio|técnica|comportamiento|escuela|creencia|razonamiento|filosofía',
	'persona' => 'persona|hombre|mujer|partidario|estilo|especialista|médico|doctor|poeta|escritor|músico|pintor|trabajador|deportista',
	'enfermedad' => 'enfermedad|trastorno|dolencia|patología|afección|inflamación|lesión|infección',
	);

	%{$repel{spanish}} = (
	# These are repulsion rules.
	# They state, for instance, that morphemes of properties
	# cannot be found in entities, and so on.
	# pending: this rules should be automatically inferred from 
	# the coreont
	'propiedad' => 'entidad',
	'proceso' => 'entidad',
	'enfermedad' => 'persona|entidad|ciencia|técnica|tratamiento|instrumento',
	'ciencia' => 'entidad',
	'técnica' => 'entidad',
	'tratamiento' => 'entidad',
	'máquina' => 'propiedad|eventualidad',
	);


	# we eliminate certain categories form the taxonomy
	# pending: this too should be read from the coreont
	$killhiper{spanish} = 'período|siglo|tiempo|momento|hora|mes|semana|día|noche'.
	'|leche|sangre|carne|huevo'.
	'|cuello|ano|corazón|piel|mano|ojo|cabeza|dedo|brazo|pie|perna|boca'.#  bodyparts
	'|lado|punta|extremo|centro|exterior|superficie|hoja|línea'.
	'|hijo|padre|madre|hermano|abuelo'.# kindship terms
	'|variedad|manera|modo'.# etc
	'|grupo|tipo|familia|especie'.
	'temperatura||guerra|empresa'.
	'|habitante|ciudad|departamento'.
	'|elemento|aire|mar|fuego|fuerza|vida|realidad|mundo|prnl'.
	'|asunto|tema|sentido|causa|importancia|diferencia'.
	'|palabra|giro|expresión';



	# ~~~~~~~~~~ French rules ~~~~~~~~~~~~~~~~~

	%{$morfemas{french}} = (
	'action' => 'ment|[ntc]ion|age',
	'propiété' => 'ité|ence',
	'science' => 'ique|logie|aphie|étrie',
	'technique' => 'graphie|ing',
	'traitement' => 'thérapie',
	'chirurgie' => 'chirurgie|tomie|plastie',
	'instrument' => 'oir|otome|scope|grahe|mètre',
	'phobie' => 'phobie',
	'doctrine' => '[ïi]sme',
	'personne' => 'ien|[ea]nt|eux|éiste|cide|phile|phobe|phone|aire|pathe|[io]mane|logue', 
	'maladie' => 'pathie|triteosis|[oa]sis|itis|asmie|oma',
	);

	%{$exceptuando{french}} = (
	'doctrine' => 'organisme|micro-organisme',
	
	);

	%{$conceptos{french}} = (
	'entité' => 'personne|homme|humain|femme|animal|substance|objet|plante',
	'action' => 'action|événement|état|activité|coup|fait|guerre|repas|sentiment|formation|changement|operation|dessin|accord|procédure|division|paiement|altération|impression|violence|céremonie|défaut|maladie|trouble|production|fabrication|augmentation|réaction|génération',
	'propiété' => 'propiété|qualité|état|manque|caractère|condition|sentiment|capacité|difficulté|volonté|circonstance|comportament|tristesse',
	'science' => 'étude|science|connaissance|médecine|philosophie',
	'technique' => 'art|norme|technique|système|diagnostic|étude|méthode',
	'traitement' => 'thérapie|traitement',
	'instrument' => 'instrument|appareil|machine',
	'phobie' => 'phobie|peur|crainte',
	'doctrine' => 'art|doctrine|idéologie|science|style|étude|théorie|mouvemeent|système|maladie|état|activité|attitude|tendance|sport|excercice|technique|comportement|école|croyance|raisonnement|philosophie',
	'personne' => 'personne|homme|femme|partisan|style|spécialiste|médecin|docteur|poète|écrivain|musicien|peintre|travailleur|sportif',
	'maladie' => 'maladie|trouble|pathologie|affection|inflammation|blessure|infection',
	);

	%{$repel{french}} = (
	'propiété' => 'entité',
	'action' => 'entité',
	'maladie' => 'personne|entité|ciencia|technique|traitement|instrument',
	'ciencia' => 'entité',
	'technique' => 'entité',
	'traitement' => 'entité',
	);

	$killhiper{french} = 'feuille|peau|pied|tête|un|an|sang|cheveu|parole|poil|branche|type|morceau';

	# ~~~~~~~~~~ English rules ~~~~~~~~~~~~~~~~~

	%{$morfemas{english}} = (
	'action' => 'age|ment|\-up|out|down|sal|ittal|[st]ure|very|ffle|cide|[st]ion|ada|azgo|[nce]ea',
	'property' => 'ment|ness|ity|sion|[ea]nce|ncy',
	'science' => 'ics|logy|omy|graphy|[es]try|mics',
	'treatment' => 'therapy|scopy',
	'surgery' => 'surgery|ctomy',
	'instrument' => 'graph|scope|phone|meter',
	'phobia' => 'phobia',
	'fish' => 'fish',
	'doctrine' => 'ism',
	'person' => 'brain|eer|rder|ndent|ager|grapher|[nl]iac|ist|sser|rter|wit|person|rrant|man|[tr]ee|path|ician|[dtc]ian', 
	'disease' => '[ao]sis|itis|pathy|emia|oma',
	'sausage' => 'wurst',
	);

	%{$exceptuando{english}} = (
	'doctrine' => 'organism|micro-organism',
	'person' => 'career|beer',
	);

	%{$conceptos{english}} = (
	'entity' => 'person|man|woman|animal|substance|object|plant',
	'action' => 'act|process|action|state|activity|fact|war|sentiment|feeling|development|change|felony|crime|surgery|operation|damage|drawing|agreement|formation|procedure|division|payment|alteration|impresion|violence|ceremony|defect|disease|disorder|production|increase|reaction|generation',
	'property' => 'property|quality|state|lack|character|condition|sentiment|feeling|capacity|difficulty|will|circunstance|behaviou?r|defect|sadness',
	'science' => 'science|knowledge|study|medicine|philosophy',
	'technique' => 'art|norm|technique|system|prognosis|study|method',
	'treatment' => 'therapy|treatment',
	'instrument' => 'instrument|apparatus|device|machine',
	'phobia' => 'phobia|fear',
	'doctrine' => 'art|doctrine|ideology|science|style|study|theory|movement|system|disease|word|state|activity|attitud|tendency|sport|exercise|technique|behaviou?r|school|belief|philosophy',
	'person' => 'person|man|woman|advocate|style|specialist|doctor|poet|writer|musician|painter|worker|sportsman',
	'disease' => 'disease|disorder|pathology|affliction|inflammation|infection',
	);

	%{$repel{english}} = (
	'property' => 'entity',
	'action' => 'entity',
	'disease' => 'person|entity|science|technique|treatment|instrument',
	'science' => 'entity',
	'technique' => 'entity',
	'treatment' => 'entity',
	);

	$killhiper{english} = 'element|name|skin|branch|head|hair|way|eye|leaf|foot|end|specie|leg|brain|class|type|stomach|variety|heart';

}

sub morfrules {
	return if (!$mods{morfrules});
	my ($hipo) = @_;
	return if (!$hipo);
	print "\n<br>morfrules dispatched with ($hipo)" if ($vbs);
	foreach my $morf (sort keys %{$morfemas{$lang}}) {
		my $check = $morfemas{$lang}{$morf};
		if ($hipo =~ /...($check)$/) { 
			# at least three letters before
			my $comment = "I detect $morf ($1) in $hipo";
			if ($hotkeys{$morf}) {
				$rank{$morf} += 1000;
				print "\n<br>$comment: $rank{$morf} " if ($vbs);
				return;
			} else {
				my $ok = &tothetop($morf, "__filler__", "hotkeys", "morfrules");
				if ($hotkeys{$ok}) {
					$rank{$ok} += 1000;
					print "\n<br>Morfrlues says that $morf takes back to $ok: $rank{$ok}" if ($vbs);
				} elsif ($ok) {
					print "\n<br>Morfrlues says $morf takes to $ok, which is not in hotkeys."  if ($vbs);
				}
			}
		}
	}
	return 1; 
}

sub asymmetric {
	return if (!$mods{asymmetric});
	print  "\n<br>asymmetric is called..." if ($vbs);
	my $in = $_[0];	
	my $rank;
	my %myrank; 
	my %disp;
	foreach my $feat (sort{$termfeat{paradigm}{$in}{$b}<=>$termfeat{paradigm}{$in}{$a}} keys %{$termfeat{paradigm}{$in}}){
		next if ($feat eq $in);
		$rank++;
		last if ($rank > 15 || $termfeat{paradigm}{$in}{$feat} < 4); 
		# pending: these should be params
		$disp{$feat}++;
		if (!$termfeat{paradigm}{$feat}) {
			print "\n<br>asymmetric is loading vector <i>$feat</i>..." if ($vbs);
			&vectorload($feat); 
			print "\n Ok. " if ($vbs); 
		}
		my $minirank;
		foreach my $featdelfeat (sort{$termfeat{paradigm}{$feat}{$b}<=>$termfeat{paradigm}{$feat}{$a}} keys %{$termfeat{paradigm}{$feat}}){
			next if ($featdelfeat eq $feat || $featdelfeat eq $in);
			$minirank++;
			last if ($minirank > 15  || $termfeat{paradigm}{$in}{$featdelfeat} < 4);
			$disp{$featdelfeat}++;
			if ($hotkeys{$featdelfeat}) {
				my $suma = 17 - $minirank; 
				# pending: these should be params
				print "\n<h2>asymmetric: {$in}->{$feat}->{$featdelfeat}->$suma</h2>"  if ($vbs);
				$myrank{$featdelfeat} += $suma;
			}
		}
	}
	undef $rank;
	foreach my $f (sort { $myrank{$b} <=> $myrank{$a} } keys %myrank ) {
		$rank++;
		last if ($rank > 15) ; 
		# pending: this should be a param
		my $vale = ($myrank{$f} * ( 1 + $disp{$f} ));
		if ($vale) {
			$rank{$f} += $vale;
			$consider{$f} += $vale;
			print "\n<h2>Asymmetric says: {$in}->{$f}->{$vale}->{$rank{$f}} ($myrank{$f} * ( 1 + $disp{$f} )}</h2>" if ($vbs);
		}
	}
}

sub contextos {
	# pending: change the name of this sub 
	# so we have all in one language (English)
	my $d = $_[0];
	my $file = $contexts."/".$d; 
	if (! -e $file ) {
		print "\n<br>Searching for contexts of <i>$d</i> in the corpus...";		
		&kwicall($d);
	}
	if (! -e $file ) {
		print "\n<br>I could not find contexts of $d.";
		return;
	}
	# pending: use the readfile sub
	open( my $fh, '<', $file ) or print "\n<br>I can't read [$file]" and return;
	my @contextos = <$fh>;
	close $fh;
	return @contextos;
}

sub sintagm {
	return if (!$mods{sintagm});
	my ($one,$two) = @_;
	foreach my $case ($one,$two) {
		if (!$contextos{$case}) {
			@{$contextos{$case}} = &contextos($case);
		}
	}
	my $c; # matches
	my $tot; # total lines 
	foreach my $l (@{$contextos{$one}}) {
		last if ($c > 5000);
		$l = lc $l; 
		$l = " ".$l." ";
		if ($l =~ /[\/ ]$two[\/ ]/ ) {
			$c++;
			# Hit of $two in $one
		}
		$tot++;
	}
	if ($c && $tot) {
		my $score = ($c / $tot);
		if ($score >= $umbrales{sintagm}) {
			return ($score, $c, $tot);
		}
	} else {
		return (0, $c, $tot);
	}
}


sub palex {
	return if (!$mods{palex});

=pod
This is the palex module: a quantitative implementation of lexical patterns

The use of lexical patterns as a means to extract hypernymy relations from corpus dates back to the work of \cite{Hearst:1992}, but is rooted in work previously undertaken on dictionary definitions \cite{Chodorow85}.
A lexical pattern is a sequence such as \textit{is a type of}. In her original paper, Hearst \cite{Hearst:1992} suggested searching for lexical patterns such as these in order to extract hypernymy relations between the nouns or noun phrases found at each side of the pattern, and many researchers used this approach.
However, lexical patterns applied in this way tend to be error prone. It is in this context
that 
efforts were made to try to identify which patterns are more reliable \cite{Potrich2008,Seitner2016}.

We believe that instead of focusing energy on
trying to select a number of patterns, better results could be achieved by a simple change in tactics. Instead of starting with the pattern, i.e. instead of looking up a given pattern in corpus, 
we propose starting by looking
up a given input noun in corpus. Once a large number of contexts of occurrence of a given noun have been extracted, 
then we suggest searching for
the occurrence of lexical patterns within these contexts. These patterns will yield a number of hypernym candidates, and the key of this procedure is precisely the frequency of occurrence of such candidates. There will be one hypernym candidate with the highest frequency, and if the frequency is not high enough (three times at least) then the candidate is rejected and the module does not return any result. This means that results will be produced only on some occasions, but when they are, they will have a high probability of being correct.

The following are some examples of the patterns we compiled for English, and similar patterns were used for Spanish and French. 

\begin{verbatim}
HYPO is a HYPER
HYPO is a type of HYPER
HYPO, a kind of HYPER
HYPO and other HYPER
HYPER such as HYPO
...
\end{verbatim}

For economy, many of these patterns are coalesced using regular expression syntax, such that one line of code can include many patterns, like this:
\begin{verbatim}
is a (type|form|kind) of
\end{verbatim}

###############
# This module has its own configuration
###############


# Now the action begins
###############
=cut

	undef %candidatos;
	undef %tree;
	undef $fondo;
	undef %examp;
	my ($input, $ref) = @_; 
	# it is called with a reference to the hash of contexts
	my @contextos;
	if ($ref) {
		@contextos = @{$ref};
	}
	my $frec = scalar(@contextos);
	if (!$ref || !$frec) {
		@contextos = &contextos($input);
	}
	$frec = scalar(@contextos);
	if (!$frec) {
		return;
	}
	print "\n<br/> palex dispatched with params: ($input, $lang)" if ($vbs > 1);
	my @utp = split /\n/, $data;
	my $hitch = ""; 
	my %return;
	my %tree;

	foreach my $ut (@utp) {
		$ut =~ s/^(hyper|hypo)\t//;
		my $tipo = $1;
		$ut =~ s/ /"}{"/g;
		next if (!$ut);
		my $code = '$tree{"'.$ut.'"} = "'.$tipo.'"';
		eval $code;
	}
	undef $fondo;
	print "\n<br>Calling parser with ".(scalar (@contextos)). " contexts." if ($vbs > 1);
	while ( my $line = shift @contextos ) {
		my $templine ; 
		my @tmp = split / +/, $line;
		foreach my $t (@tmp) {
			$t = lc ($t);
			my @ttmmpp = split /\//, $t;
			$templine .= $ttmmpp[0]." ";
			$vocab{$ttmmpp[0]}++;
		}

		chomp $templine;

		foreach my $num (0 .. $#regex) {
			# pending: this is suboptimal from the efficiency
			# point of view
			# Something better would be to join the regex:
			# $reglas = join '|', @regex;
			# and then:
			# if ($line =~ /$reglas/ )
			my $regex = $regex[$num];
			if ($templine =~ /($regex)/ ) {
				my $catch = $1;
 				print "\n\n<hr><br/>$templine\n<br/>Found:<i>$catch</i><br/>Rule ".($num+1)  if ($vbs > 1);
				&parser($line, $input, $templine);
 			}
		}
	}

	# we select the best candidate:
	my @sort = (sort {$candidatos{$b} <=> $candidatos{$a}} keys %candidatos);
	my $best;
	while (!$best && scalar @sort) {
		$best = shift @sort;
		$best = &cleant($best);
	}
	while ($best =~ /$input/ 
		# the hyponym cannot be included in the hypernym
		|| $best =~ /^(acepción|término|nombre) /
		# pending: this should be in the stoplist
		|| ($input =~ /$best/ && $input !~ /^$best/ && $lang =~ /spanish|french/) 
		# if there is overlap, the hypernym should be first
		# in this language pair
		|| ($input =~ /$best/ && $input !~ /$best$/ && $lang eq "english") 
		# or at the end, en English
		) {
		$best = shift @sort;
	}
	return if ($candidatos{$best} < 3); 
	# pending: this should be a param 
	my $scorcese = ($candidatos{$best}/$frec) ;
	my $limit = 0.01;
	# pending: this too is a param 
	if ($scorcese > $limit) { 
		my $ret = "($candidatos{$best}/$frec)=$scorcese > $limit";
		$best =~ s/^ +//g;
		my $lema = &noplur($best);
		$best = $lema if ($lema);
		$best =~ s/^(forma|tipo|clase|especie|uso|utilización|función) del? //g;
		
		# pending: sometimes it gets things like these:
		# antiguo máquina de labranza
		# diferente actor
		# pequeño maquinaria agrícola
		$best =~ s/^(antiguo|pequeño|diferente) //g;
		# pending: again this should be in the stoplist		

		return ($best, $candidatos{$best});
		# pending: it should send the contexts too
	}
}


sub parser {
	# not a very good name, for what it actually does.
	# It receives a sequence and analyzed it searching for the patterns
	my (%lema, %pos);
	my $return = "";
	my $tagged = $_[0];
	my $input = $_[1];
	my $templine = $_[2];
	$templine =~ s/###/<b>/;
	$templine =~ s/###/<\/b>/;
	$tagged =~ s/\t###\t(.+)\t###\t/ $1 /;
	my @tmp = split / +/, $tagged;
	# The corpus is tagged, and so we convert it to a hash
	my $line = "";
	foreach my $t (@tmp) {
		# pending: there is a double loop. 
		# There should be only one.
		$t = lc ($t);
		my @ttmmpp = split /\//, $t;
		$t = $ttmmpp[2]; # we convert forms to lemmas
		$t = $ttmmpp[0] if ($t eq "<unknown>");
		$line .= $t." ";
		$lema{$ttmmpp[0]}{$t}++;
		$pos{$t}{$ttmmpp[1]}++;
	}
	# We often encounter bibliographical references and we have
	# to get rid of them
	if ($line =~ /$ref/) {
		$line =~ s/$ref//g;
	}
	$line =~ s/^ *([0-9]+)//;
	my $details = "\n<br><i>$line</i>";
	
	my ($flush, $sintagma, $enumera, @cohypo);
	my $boost; # this will contain a NP found before a parenthesis that contains the target noun
	my $boostnext; # this tells if a patters is found
	my @trayecto;
	my $surebet; # a sure bet is found when it has a clear pattern
	my $lastone; # last NP detected
	foreach my $position (0 .. $#tmp) {		
		my $tok = $tmp[$position];	 
		next if (!$tok); 
		if ($tok eq $input
			 && $tmp[($position+1)] eq "be"
			 && ($tmp[($position+2)] eq "a" ||  $tmp[($position+2)] eq "the" )
			# pending: we need order here
			) {
			 $boostnext++;
			$details .= "\n<br>Pattern detected! Hyper is comming!";
		}
		my @pos = sort { $pos{$tok}{$b} <=> $pos{$tok}{$a} } keys %{$pos{$tok}};
		my $pos = $pos[0];

		if ($tree{$tok}) { 
			# First element of pattern detected 
			if ($tree{$tok} =~ /(hypo|hyper)/) {
				# Full pattern detected
				$surebet = $tree{$tok};
			} else {
				push (@trayecto, $tok);
			}
		} elsif (scalar (@trayecto)) {
			push (@trayecto, $tok);
			my $save;
			my $path = join '"}{"', @trayecto;
			my $code = '$save = ($tree{"'.$path.'"})';
			$details .= "\n<br>Code: $code";
			eval $code;
			if ($save) {
				# Partial pattern detected
				if ($save =~ /(hypo|hyper)/) {
					$details .= "\n<br>Full pattern detected: $save {\"$path\"}";
					$surebet = $save;
				}
			} else  { # no pattern was found
				undef @trayecto; 
			}
		}

		if ($pos =~ /^($noun)$/  && !$sintagma && !$kill{$tok}) { 
			# first part of an NP
			$sintagma = $tok." ";
			undef $flush;
		} elsif ($sintagma && ($tok =~ /^ *($middle) *$/ 
			|| ($pos =~ /^($noun)$/ && !$kill{$tok} ))) { 
			$sintagma .= $tok." ";
			# The NP continues
		} elsif ($sintagma) {
			# The NP finishes here
			$sintagma =~ s/ +$//g;
			if (length($sintagma) < 3 
			|| length($sintagma) > 65) {
				# pending: these numbers are so arbitrary...
				undef $sintagma;
			} else {
				$sintagma =~ s/ ($middle)$//g;
				# NP ends here
				if (
				( $tmp[($position+1)] eq "such"
				&& $tmp[($position+2)] eq "as"
				&& $tmp[($position+3)] eq $input )
				|| ( ($tmp[($position+1)] eq "like" || $tmp[($position+1)] eq "(")
				&& $tmp[($position+2)] eq $input ) 
				|| ( $tmp[($position+1)] eq "in"
				&& $tmp[($position+2)] eq "particular"
				&& $tmp[($position+3)] eq $input )
				) {
					$surebet = "hyper";
				}

				if ($surebet eq "hyper") {
					$candidatos{$sintagma} += 5;
					$boosty{$sintagma}++;
					undef $surebet;
				} elsif ($surebet eq "hypo" && $sintagma eq $input ) {
					$candidatos{$boost} += 5;
					$boosty{$boost}++;
					undef $surebet;
					undef $boost;
				}

				if ($sintagma eq $input || $sintagma =~ /$input/) {
					undef $sintagma;
					next;
				} 
				if (!$examp{$sintagma}) {
					$examp{$sintagma}{$templine}++;
				}
				if (!$enumera) { 
					# If we are inside an enumeration
					# we can't be hypernyms
					$candidatos{$sintagma}++;
					if ($boostnext) { 
						$candidatos{$sintagma} += 5;
						$boosty{$sintagma}++;
						undef $boostnext;
					}
				} else {
					push (@cohypo, $sintagma); 
				}
			}
			$flush++; 
			# this indicates that the NP is closed
		}

		if ($tok eq "(") {
			# An enumeration
			if ($sintagma || $lastone) { 
				$boost = $sintagma; 
				# it is stored here until it is 
				# confirmed that the parenthesis contains
				# the target noun
				$boost = $lastone if (!$boost);
				# Parenthesis triggers buffer $boost]
				undef $sintagma;
				undef $lastone;
			}
			$enumera++;
		} elsif ($tok eq ")" || ($enumera && $position == $#tmp)) {
			undef $enumera;
			# Parenthesis is closed
			if (scalar(@cohypo) ) {
				# It checks if the target is among the 
				# enumerated elements 
		   		if ($boost) {
					$candidatos{$boost} += 5;
		 			$details .= "\n<br>Boost:[$boost] ~ $candidatos{$boost}";
					undef $boost;
		   		}	
				foreach my $co (@cohypo) {
					next if ($co eq $input || $kill{$co}); 
					$cohyperonyms{$co}++;
				}
				undef @cohypo;
			}
		}

		if ($flush) {
			$lastone = $sintagma;
			undef $sintagma;
		}
	}
	print $details if ($vbs > 1);
	return $return;
}



##############
# Module subroutines end here
##############


##############
# Auxiliary subroutines start here
##############

sub readfile {
	# to read a file
	my $file = $_[0];
	if (! -e $file) {
		print "\n<br>File not found: [$file]" and return;
	}
	open(my $t, '<', $file) or print "\n<br>I can't read file [$file]: $? $!" and return;
	my @file = <$t>;
	close $t;
	return @file;
}

sub vectorate {
	# This one is used to create co-occurrence vectors
	# from the contexts of occurrence.
	# Once created, they are stored in disk so they don't
	# have to be created again every time.
	my $d = $_[0];
	my %frec;
	my $output = $vecfold."/".$d;
	if (-e $output) {
		print "\n<br>Vector [$d] already exists." if ($vbs);
		return;
	}
	if (!$contextos{$d}) { 
		@{$contextos{$d}} = &contextos($d);
	}

	my $c;
	foreach my $line (@{$contextos{$d}}) {
		$c++;
		last if ($c > 5000);
		$line =~ s/\r?\n//g;
		while ( $line =~ s/([^\/ \t]+)\/(ADJ|JJ|[NV])[^\/ \t]*\/([^\/ \t]+)/ / ) {
			my $form = $1;
			$vocab{$form}++;
			my $pos = $2;
			$pos = "ADJ" if $pos eq "JJ";
			my $l = $3;
			$l = $1 if $l eq "<unknown>";
			$l = lc $l;
			$frec{$l."\t".$pos}++;
		}
	}

	my $thisfile = "";
	foreach my $w (sort {$frec{$b} <=> $frec{$a} } keys %frec){
		last if ($frec{$w} < 3);
		$thisfile .= "\n$w\t$frec{$w}";
	}
	open(OUT, ">".$output) 
		or print "\n<br>I can't create $output: $!" and die;
		# one can't just die in a web environment.
	print OUT $thisfile;
	close OUT;
}

sub head {
	my $in = $_[0];
	if ($in =~ /. ./) { 
		if ($lang eq "english") {
			# in English, it's the last component
			$in =~ s/ of .+//g; 			
			$in =~ s/^.+ //g; 
		} else { 
			# pending: this will have to be adjusted when 
			# more languages are included
			$in =~ s/ .+//g; 
		}
		return $in;
	}
}

sub noplur {
	my $in = $_[0];
	my @p = split / /, $in;
	my $out = "";
	foreach my $trozo (@p) {
		my $lema = &noplursing($trozo);
		if ($lema ne $trozo) {
			$out .= $lema." ";
		} else {
			$out .= $trozo." ";
		}
	}
	$out =~ s/ +$//g;
	$out =~ s/  +/ /g;
	return $out;
}

sub noplursing {
	# this was taken from Pullpos http://www.tecling.com/pullpos
	# pending: find a better way
	my $in = $_[0];
	my $nonplurals = '(us|[oipeas][lrxtfsd]is|[íó][dtf][aei][dnl]es)';
	return if ($in =~ /$nonplurals$/);
	my %transfer;
	if ($lang eq "spanish") { 
		# pending: this could be done with a two dimensional hash
		%transfer = (
		'áceos' => 'áceo',
		'aces' => 'az',
		'ades' => 'ad',
		'adas' => 'ada', 
		'an[ae]s' => 'án',
		'unes' => 'ún',
		'antes' => 'ante',
		'ares' => 'ar',
		'arias' => 'aria',
		'atos' => 'ato',
		'bles'	=> 'ble',
		'cias' => 'cia',
		'dos' => 'do',
		'eanos' => 'eano',
		'eces' => 'ez',
		'ejos' => 'ejo',
		'entes' => 'ente',
		'entos' => 'ento',
		'enos' => 'eno',
		'eones'=> 'eón',
		'eos' => 'eo',
		'geas' => 'gea',
		'ian[ao]s' => 'iano',
		'ias' => 'ia',
		'ices' => 'iz',
		'íces' => 'íz',
		'ic[oa]s' => 'ico',
		'idades' => 'idad',
		'id[oa]s' => 'ido',
		'inas' => 'ina',
		'ines' => 'ín',
		'iones' => 'ión',
		'ísim[ao]s' => 'ísimo',
		'ismos' => 'ismo',
		'istas' => 'ista',
		'itos' => 'ito',
		'ivas' => 'iva',
		'iv[oa]s' => 'ivo',
		'les' => 'l',
		'manos' => 'mano',
		'mientos' => 'miento',
		'ógi[ao]s' =>  'ógico',
		'oides' => 'oide',
		'ólog[ao]s' => 'ólogo',
		'omas' => 'oma',
		'onas' => 'ona',
		'ones' => 'ón',
		'or[ea]s' => 'or',
		'os[ao]s' => 'oso',
		'rías' => 'ría',
		'rios' => 'rio',
		'ríos' => 'río',
		'sías' => 'sía',
		'stas' => 'sta',
		'stros' => 'stro',
		'tías' => 'tía',
		'udes' => 'ud',
		'uras' => 'ura',
		'úmenes' => 'umen'
		);
	} elsif ($lang eq "french") {
		# pending!
		%transfer = (
		"ents" => "ent",
		"ions" => "ion",
		'les' => 'l',
		);
	} elsif ($lang eq "english") {
		# pending!
		%transfer = (
		"ents" => "ent",	
		"ions" => "ion",
		);
	}

	my $copy = $in;
	if ($copy =~ s/s$//) {
		if ($vocab{$copy}) {
			return $copy;
		} 
	}

	if ($copy =~ s/es$//) {
		if ($vocab{$copy}) {
			return $copy;
		} 
	}

	foreach my $plu (sort keys %transfer) {
		my $copy = $in;
		my $remp = $transfer{$plu};
		if ($copy =~ s/$plu$/$remp/) {
			return $copy;
		}
	}
}

sub isincore {
	# This is to detect if a noun is in the coreont
	# and, if it is, to return the immediate hypernym
	my ($in, $isfirst) = @_;
	return if ($stop{$in} && !$isfirst); 
	# if it is the first one, it will not check the stoplist
	my @t = keys %{$upward{$in}}; 
	# at the moment we have a single hypernym per noun in the coreont
	# but we could have more than one in the future
	# what will be return then?
	# pending: figure this out
	if (scalar (@t)) {
		return $t[0];
	}
}

sub cachemira {
	# This is to see is a noun is in the cache
	my $in = $_[0];
	my $cachefile = $cachedir.$in;
	return if (!-e $cachefile);
	print "\n<br>[$in] has been classified already." if ($vbs);
	open( my $fh, '<', $cachefile ) or 
		print "\n<br>I could not open file [$cachefile] ($? $!)" 
		and return;
	my @t = <$fh>;
	close $fh;
	my $pop = pop @t; 
	# It always returns the last one. 
	# This will be a problem with polysemous terms
	# pending: figure out a better way to do this
	my @cache = split /\t/, $pop;
	shift @cache; # the first one is an empty position
	shift @cache; # the second is the term
	shift @cache; # the third is the frequency
	return @cache;
}

sub kwicall {
	# This is to make the call to kwico.
	# pending: incorporate it as module instead of 
	# calling it by the shell
	my $in = $_[0];
	my $out = $contexts."/".$in;
	foreach my $corpus (@{$corpora{$lang}}) {
		# it has to check whether the corpus exists:
		if (! -d $corpus."_index") {
			print "\n<br/>Kwico cannot find the selected corpus: $corpus";
			next;
		}
		my $order = "$kwico --max $max --annotation --corpus $corpus --query \"$in\" >> \"$out\"";
		print "\n<br/>Retrieving concordances of <i>$in</i> from $corpus..." if ($vbs); 
		print ".";
		system ($order);	
	}
	if (! -e $out) {
		print "\nError: file [$out] could not be created!";
		next;
	}
}

sub cleant { # clean term
	my $in = $_[0];
	$in =~ s/['":;\[\]\(\)\\\/@\+\*\-\{\}\_\.0-9]+//g;
	$in =~ s/ +$//g;
	$in =~ s/^ +//g;
	return $in;
}

sub savefile {
	my ($in, $chich) = @_;
	if ( -e $in) {
		print "\n<br>File [$in] already exists. Changing to:";
		$in =~ s/\.(txt|csv)//g;
		$in .= time().".txt";
		print $in;
	}
	open(LOCAL, ">$in") or print $! and die;
	print LOCAL $chich;
	close LOCAL;
	print "\n<br>Ok, data generated at: $in";
}

sub institutional {
	print "
	<h4>Abstract</h4>
	<p style='color:#111111'> 
We designed a statistically-based
taxonomy induction algorithm consisting of a combination of different strategies not involving explicit linguistic knowledge. Being all
quantitative, the strategies we present are however of different nature. Some of them are based on the computation of distributional
similarity coefficients which identify pairs of sibling words or co-hyponyms, while others are based on asymmetric co-occurrence and
identify pairs of parent-child words or hypernym-hyponym relations. A decision making process is then applied to combine the results
of the previous steps, and finally connect lexical units to a basic structure containing the most general categories of the language. 
We are currently in the process of evaluating the latest results.

 <a href='$self&source=1'>
<h4>Documentation and source code</h4>


 <p style='color:#111111'> 
Today is Friday, December 11, 2020. The source code and documentation have been changing very rapidly and we are still working on the details. However, you can have a look at what we've got so far, 
which is a stable version.

<br> We will not maintain older versions.

<br>If you would like to send inquiries
you are welcome to do so at rogelio (dot) nazar (at) gmail (dot) com .
 </a>

<h4>Funding</h4>
	<p style='color:#111111'> 
<br>This project has been supported by two successive grants: 
<ol>
<li>Conicyt-Fondecyt 11140686, “Inducción
automática de taxonomías de sustantivos generales y especializados a partir de corpus textuales desde el enfoque de
la lingüística cuantitativa” (Automatic taxonomy induction from corpora for terminology and general vocabulary using quantitative measures). Lead researcher: Rogelio Nazar. (2014 to 2017).
<li> Ecos Sud-Conicyt Project C16H02 “Inducción automática de taxonomías del español y el francés mediante técnicas cuantitativas y estadística de corpus” (Automatic taxonomy induction from corpora for Spanish and French using quantitative corpus analysis). Lead researcher: Irene Renau. (2016-2019).
</ol>

<h4>Credits</h4>
	<p style='color:#111111'> 
Researchers:
<ul>
<li>Rogelio Nazar 
<li>Antonio Balvet
<li>Gabriela Ferraro
<li>Rafael Marín 
<li>Irene Renau 
</ul>


<h4>Related publications:</h4>
	<p style='color:#111111'> 
<ul>
 <li><p><a href='https://www.degruyter.com/downloadpdf/journals/jisys/30/1/article-p376.xml'>
 <li>+ Nazar, R.; Balvet, A.; Ferraro, G.; Marín, R.; Renau, I. (2020). Pruning and repopulating a lexical taxonomy: experiments in Spanish, English and French. Journal of Intelligent Systems, vol. 30 num. 1, pp. 376-394. </a>

 <li><a href='http://journal.sepln.org/sepln/ojs/ojs/index.php/pln/article/download/6192/3643'>
+ Nazar, R.; Obreque, J.; Renau, I. (2020). Tarántula –> araña –> animal : asignación de hiperónimos de segundo nivel basada en métodos de similitud distribucional. Procesamiento del Lenguaje Natural, núm 64, pp. 29-36. </a>
 <li>+ Nazar, Rogelio. (2019). El análisis cuantitativo de la coocurrencia léxica en la lexicografía especializada. In Sanmartín Sáez, Julia y Quilis Merín, Mercedes (eds.). Retos y avances en lexicografía: los diccionarios del español en el eje de la variación lingüística. Anejo 10 de Normas. Valencia: Asociación Española de Estudios Lexicográficos.
 <li>+ Nazar, R., Renau, I., Marín, R. (2017). Experiments in taxonomy induction in Spanish and French. Proceedings of Language, Ontology, Terminology and Knowledge Structures Workshop (LOTKS 2017), pp. 66-75. 
 <li>+ Nazar, R., Renau, I., Marín, R. (2017). Taxonomía automatizada de sustantivos del castellano y del francés: hacia el etiquetado semántico automático multilingüe. In: Sariego López, Ignacio, Juan Gutiérrez Cuadrado y Cecilio Garriga Escribano (eds.), El diccionario en la encrucijada: de la sintaxis y la cultura al desafío digital. Santander: AELEX, 731-745. 
 <li>+ Renau, I., Nazar, R. (2017). Verbos en contexto: una propuesta para la detección automática de patrones léxicos en corpus. In: Sariego López, Ignacio, Juan Gutiérrez Cuadrado y Cecilio Garriga Escribano (eds.), El diccionario en la encrucijada: de la sintaxis y la cultura al desafío digital. Santander: AELEX, 879- 897. 
 <li>+ Nazar, R.; Arriagada, P. (2017). POL: un nuevo sistema para la detección y clasificación de nombres propios. Procesamiento del Lenguaje Natural, n. 58, pp. 13-20. 
 <li>+ Nazar, R.; Soto, R.; Urrejola, K. (2017). Detección automática de nombres eventivos no deverbales en castellano: un enfoque cuantitativo basado en corpus. Revista Linguamatica, vol. 9, num. 2, pp. 21-31. 
 <li>+ Nazar, R. (2016). Distributional analysis applied to terminology extraction: example in the domain of psychiatry in Spanish. Terminology: International Journal of Theoretical and Applied Issues in Specialized Communication, 22(2):142-170.
 <li>+ Nazar, R., Renau, I. (2016). A Quantitative analysis of the semantics of verb-argument structures. In S. Torner and E. Bernal (eds.), Collocations and other lexical combinations in Spanish. Theoretical and Applied approaches. New York: Routledge, pp. 114-136.
 <li>+ Nazar, R.; Renau, I. (2016). Automatic extraction of lexico-semantic patterns from corpora. Proceedings of the XVII EURALEX International Congress: Lexicography and Linguistic Diversity. Tinatin Margalitadze and George Meladze (eds). Tbilisi, Gergia: Ivane Javakhishvili Tbilisi State University, pp. 823-830. 
 <li>+ Nazar, R.; Renau, I. (2016). A taxonomy of Spanish nouns, a statistical algorithm to generate it and its implementation in open source code. Proceedings of the 10th International Conference on Language Resources and Evaluation (LREC'16). European Language Resources Association (ELRA), May 2016, pp. 1485-1492. 
 <li>+ Nazar, R.; Renau, I. (2015). Agrupación semántica de sustantivos basada en similitud distribucional: implicaciones lexicográficas. In María Pilar Garcés Gómez (ed.): Lingüística y diccionarios (Anexos Revista de Lexicografía, vol. 2: 272-295). Universidade da Coruña.
 <li>+ Nazar, R.; Renau, I. (2015). Ontology Population Using Corpus Statistics. Proceedings of the Joint Ontology Workshops 2015 co-located with the 24th International Joint Conference on Artificial Intelligence (IJCAI 2015). Buenos Aires, Argentina, July 25-27, 2015. 

<h4>Related concepts:</h4> corpus statistics, distributional semantics, Spanish, taxonomy induction

<br><br>
</td><td>&nbsp</td></tr>
</table>";

}

